import s2_data
import re
import os
from unittest import TestCase, main
from itertools import chain


dict_key_regex = re.compile('^[a-z][a-z0-9_]*$')
save_game_path = os.path.join(
    os.path.dirname(__file__),
    '../../data/saves/mostly-full-savegame.sav')

class DescriptorsTestCase(TestCase):

    def test_category_key_names(self):
        for key in s2_data.field_descriptors.keys():
            self.assertTrue(dict_key_regex.match(key))

    def test_descriptor_key_names(self):
        for category in s2_data.field_descriptors.values():
            for key in category.fields.keys():
                self.assertTrue(dict_key_regex.match(key), key)

    def test_descriptors_dont_overlap(self):
        """Fields should never share bytes."""
        categories = s2_data.field_descriptors.values()
        descriptors = list(chain(*[c.fields.values() for c in categories]))
        # Order by offset so we only have to check whether a field overlaps
        # with the following field.
        by_offset = sorted(descriptors, key=lambda d: d.offset)

        current = by_offset[0]
        for next_ in by_offset[1:]:
            with self.subTest(current=current, next_=next_):
                current_end = current.offset + current.type.size
                self.assertLessEqual(current_end, next_.offset)

                current = next_

    def test_writing_current_value_is_noop(self):
        """Converting a field's bytes to a value and back into bytes should
        always succeed and the final bytes should be the same as the original
        bytes."""
        with open(save_game_path, 'rb') as save_file:
            save = save_file.read()

            for category in s2_data.field_descriptors.values():
                for field_descriptor in category.fields.values():
                    with self.subTest(category=category.name, field=field_descriptor.name):
                        field_start = field_descriptor.offset
                        field_end = field_start + field_descriptor.type.size
                        field_bytes = save[field_start:field_end]
                        value = field_descriptor.type.from_binary(field_bytes)
                        bytes_ = field_descriptor.type.to_binary(value)
                        self.assertEqual(field_bytes, bytes_)


if __name__ == '__main__':
    main()



from dataclasses import dataclass
from enum import IntFlag, IntEnum
import typing
from .field_types import *

class CharacterUnlockFlags(IntFlag):
    Ana        = 1 << 0
    Margaret   = 1 << 1
    Colin      = 1 << 2
    Roffy      = 1 << 3
    Alto       = 1 << 4
    Liz        = 1 << 5
    Nekka      = 1 << 6
    LISE       = 1 << 7
    Coco       = 1 << 8
    Manfred    = 1 << 9
    Jay        = 1 << 10
    Tina       = 1 << 11
    Valerie    = 1 << 12
    Au         = 1 << 13
    Demi       = 1 << 14
    Pilot      = 1 << 15
    Airyn      = 1 << 16
    Dirk       = 1 << 17
    Guy        = 1 << 18
    ClassicGuy = 1 << 19

class ShortcutProgressEnum(IntEnum):
    NotMet = 0x0
    Met = 0x1
    S1d4Gave2000 = 0x2
    S1d4GaveBomb = 0x3
    S1d4Gave10000 = 0x4
    S3d1GaveRope = 0x5
    S3d1GaveWeapon = 0x6
    S3d1GaveMount = 0x7
    S5d1Gave50000 = 0x8
    S5d1GaveHiredHand = 0x9
    S5d1GaveGoldenKey = 0xa

class CharacterSelectedEnum(IntEnum):
    Ana        = 0x00
    Margaret   = 0x01
    Colin      = 0x02
    Roffy      = 0x03
    Alto       = 0x04
    Liz        = 0x05
    Nekka      = 0x06
    LISE       = 0x07
    Coco       = 0x08
    Manfred    = 0x09
    Jay        = 0x0a
    Tina       = 0x0b
    Valerie    = 0x0c
    Au         = 0x0d
    Demi       = 0x0e
    Pilot      = 0x0f
    Airyn      = 0x10
    Dirk       = 0x11
    Guy        = 0x12
    ClassicGuy = 0x13

@dataclass(frozen = True)
class FieldDescriptor:
    offset: int
    name: str
    type: Type

@dataclass(frozen = True)
class Category:
    name: str
    fields: typing.Mapping[str, FieldDescriptor]

# The dictionary keys are meant to stay consistent so that programs that use
# those keys don't break. The names in the descriptors/category can change.

field_descriptors = {
    'magic_numbers': Category('Magic Numbers', {
        'magic_num1': FieldDescriptor(0x0, 'Magic number 1', ByteType()),
        'magic_num2': FieldDescriptor(0x1, 'Magic number 2', ByteType()),
    }),
    'places_discovered': Category('Journal: Places Discovered', {
        'dwelling':         FieldDescriptor(0x02, "Dwelling",          BoolType()),
        'jungle':           FieldDescriptor(0x03, "Jungle",            BoolType()),
        'volcana':          FieldDescriptor(0x04, "Volcana",           BoolType()),
        'olmec_lair':       FieldDescriptor(0x05, "Olmec's Lair",      BoolType()),
        'tide_pool':        FieldDescriptor(0x06, "Tide Pool",         BoolType()),
        'abzu':             FieldDescriptor(0x07, "Abzu",              BoolType()),
        'temple_of_anubis': FieldDescriptor(0x08, "Temple of Anubis",  BoolType()),
        'the_city_of_gold': FieldDescriptor(0x09, "The City of Gold",  BoolType()),
        'duat':             FieldDescriptor(0x0a, "Duat",              BoolType()),
        'ice_caves':        FieldDescriptor(0x0b, "Ice Caves",         BoolType()),
        'neo_babylon':      FieldDescriptor(0x0c, "Neo Babylon",       BoolType()),
        'tiamat_throne':    FieldDescriptor(0x0d, "Tiamat's Throne",   BoolType()),
        'sunken_city':      FieldDescriptor(0x0e, "Sunken City",       BoolType()),
        'eggplant_world':   FieldDescriptor(0x0f, "Eggplant World",    BoolType()),
        'hundun_hideaway':  FieldDescriptor(0x10, "Hundun's Hideaway", BoolType()),
        'cosmic_ocean':     FieldDescriptor(0x11, "Cosmic Ocean",      BoolType()),
    }),
    'bestiary_discovered': Category('Journal: Bestiary Discovered', {
        'snake':             FieldDescriptor(0x12, "Snake",             BoolType()),
        'spider':            FieldDescriptor(0x13, "Spider",            BoolType()),
        'bat':               FieldDescriptor(0x14, "Bat",               BoolType()),
        'caveman':           FieldDescriptor(0x15, "Caveman",           BoolType()),
        'skeleton':          FieldDescriptor(0x16, "Skeleton",          BoolType()),
        'horned_lizard':     FieldDescriptor(0x17, "Horned Lizard",     BoolType()),
        'moles':             FieldDescriptor(0x18, "Moles",             BoolType()),
        'quillback':         FieldDescriptor(0x19, "Quillback",         BoolType()),
        'mantrap':           FieldDescriptor(0x1a, "Mantrap",           BoolType()),
        'tiki_man':          FieldDescriptor(0x1b, "Tiki Man",          BoolType()),
        'witch_doctor':      FieldDescriptor(0x1c, "Witch Doctor",      BoolType()),
        'mosquito':          FieldDescriptor(0x1d, "Mosquito",          BoolType()),
        'monkey':            FieldDescriptor(0x1e, "Monkey",            BoolType()),
        'hang_spider':       FieldDescriptor(0x1f, "Hang Spider",       BoolType()),
        'giant_spider':      FieldDescriptor(0x20, "Giant Spider",      BoolType()),
        'magmar':            FieldDescriptor(0x21, "Magmar",            BoolType()),
        'robot':             FieldDescriptor(0x22, "Robot",             BoolType()),
        'fire_bug':          FieldDescriptor(0x23, "Fire Bug",          BoolType()),
        'imp':               FieldDescriptor(0x24, "Imp",               BoolType()),
        'lavamander':        FieldDescriptor(0x25, "Lavamander",        BoolType()),
        'vampire':           FieldDescriptor(0x26, "Vampire",           BoolType()),
        'vlad':              FieldDescriptor(0x27, "Vlad",              BoolType()),
        'olmec':             FieldDescriptor(0x28, "Olmec",             BoolType()),
        'jiangshi':          FieldDescriptor(0x29, "Jiangshi",          BoolType()),
        'jiangshi_assassin': FieldDescriptor(0x2a, "Jiangshi Assassin", BoolType()),
        'flying_fish':       FieldDescriptor(0x2b, "Flying Fish",       BoolType()),
        'octopy':            FieldDescriptor(0x2c, "Octopy",            BoolType()),
        'hermit_crab':       FieldDescriptor(0x2d, "Hermit Crab",       BoolType()),
        'pangxie':           FieldDescriptor(0x2e, "Pangxie",           BoolType()),
        'great_humphead':    FieldDescriptor(0x2f, "Great Humphead",    BoolType()),
        'kingu':             FieldDescriptor(0x30, "Kingu",             BoolType()),
        'crocman':           FieldDescriptor(0x31, "Crocman",           BoolType()),
        'cobra':             FieldDescriptor(0x32, "Cobra",             BoolType()),
        'mummy':             FieldDescriptor(0x33, "Mummy",             BoolType()),
        'sorceress':         FieldDescriptor(0x34, "Sorceress",         BoolType()),
        'cat_mummy':         FieldDescriptor(0x35, "Cat Mummy",         BoolType()),
        'necromancer':       FieldDescriptor(0x36, "Necromancer",       BoolType()),
        'anubis':            FieldDescriptor(0x37, "Anubis",            BoolType()),
        'ammit':             FieldDescriptor(0x38, "Ammit",             BoolType()),
        'apep':              FieldDescriptor(0x39, "Apep",              BoolType()),
        'anubis_ii':         FieldDescriptor(0x3a, "Anubis II",         BoolType()),
        'osiris':            FieldDescriptor(0x3b, "Osiris",            BoolType()),
        'ufo':               FieldDescriptor(0x3c, "Ufo",               BoolType()),
        'alien':             FieldDescriptor(0x3d, "Alien",             BoolType()),
        'yeti':              FieldDescriptor(0x3e, "Yeti",              BoolType()),
        'yeti_king':         FieldDescriptor(0x3f, "Yeti King",         BoolType()),
        'yeti_queen':        FieldDescriptor(0x40, "Yeti Queen",        BoolType()),
        'lahamu':            FieldDescriptor(0x41, "Lahamu",            BoolType()),
        'proto_shopkeeper':  FieldDescriptor(0x42, "Proto Shopkeeper",  BoolType()),
        'olmite':            FieldDescriptor(0x43, "Olmite",            BoolType()),
        'lamassu':           FieldDescriptor(0x44, "Lamassu",           BoolType()),
        'tiamat':            FieldDescriptor(0x45, "Tiamat",            BoolType()),
        'tadpole':           FieldDescriptor(0x46, "Tadpole",           BoolType()),
        'frog':              FieldDescriptor(0x47, "Frog",              BoolType()),
        'fire_frog':         FieldDescriptor(0x48, "Fire Frog",         BoolType()),
        'goliath_frog':      FieldDescriptor(0x49, "Goliath Frog",      BoolType()),
        'grub':              FieldDescriptor(0x4a, "Grub",              BoolType()),
        'giant_fly':         FieldDescriptor(0x4b, "Giant Fly",         BoolType()),
        'hundun':            FieldDescriptor(0x4c, "Hundun",            BoolType()),
        'eggplant_minister': FieldDescriptor(0x4d, "Eggplant Minister", BoolType()),
        'eggplup':           FieldDescriptor(0x4e, "Eggplup",           BoolType()),
        'celestial_jelly':   FieldDescriptor(0x4f, "Celestial Jelly",   BoolType()),
        'scorpion':          FieldDescriptor(0x50, "Scorpion",          BoolType()),
        'bee':               FieldDescriptor(0x51, "Bee",               BoolType()),
        'queen_bee':         FieldDescriptor(0x52, "Queen Bee",         BoolType()),
        'scarab':            FieldDescriptor(0x53, "Scarab",            BoolType()),
        'golden_monkey':     FieldDescriptor(0x54, "Golden Monkey",     BoolType()),
        'leprechaun':        FieldDescriptor(0x55, "Leprechaun",        BoolType()),
        'monty':             FieldDescriptor(0x56, "Monty",             BoolType()),
        'percy':             FieldDescriptor(0x57, "Percy",             BoolType()),
        'poochi':            FieldDescriptor(0x58, "Poochi",            BoolType()),
        'ghist':             FieldDescriptor(0x59, "Ghist",             BoolType()),
        'ghost':             FieldDescriptor(0x5a, "Ghost",             BoolType()),
        'cave_turkey':       FieldDescriptor(0x5b, "Cave Turkey",       BoolType()),
        'rock_dog':          FieldDescriptor(0x5c, "Rock Dog",          BoolType()),
        'axolotl':           FieldDescriptor(0x5d, "Axolotl",           BoolType()),
        'qilin':             FieldDescriptor(0x5e, "Qilin",             BoolType()),
        'mech_rider':        FieldDescriptor(0x5f, "Mech Rider",        BoolType()),
    }),
    'people_discovered': Category('Journal: People Discovered', {
        'ana_spelunky':       FieldDescriptor(0x60, "Ana Spelunky",       BoolType()),
        'margaret_tunnel':    FieldDescriptor(0x61, "Margaret Tunnel",    BoolType()),
        'colin_northward':    FieldDescriptor(0x62, "Colin Northward",    BoolType()),
        'roffy_sloth':        FieldDescriptor(0x63, "Roffy D. Sloth",     BoolType()),
        'alto_singh':         FieldDescriptor(0x64, "Alto Singh",         BoolType()),
        'liz_mutton':         FieldDescriptor(0x65, "Liz Mutton",         BoolType()),
        'nekka_the_eagle':    FieldDescriptor(0x66, "Nekka the Eagle",    BoolType()),
        'lise_project':       FieldDescriptor(0x67, "Lise Project",       BoolType()),
        'coco_von_diamonds':  FieldDescriptor(0x68, "Coco von Diamonds",  BoolType()),
        'manfred_tunnel':     FieldDescriptor(0x69, "Manfred Tunnel",     BoolType()),
        'little_jay':         FieldDescriptor(0x6a, "Little Jay",         BoolType()),
        'tina_flan':          FieldDescriptor(0x6b, "Tina Flan",          BoolType()),
        'valerie_crump':      FieldDescriptor(0x6c, "Valerie Crump",      BoolType()),
        'au':                 FieldDescriptor(0x6d, "Au",                 BoolType()),
        'demi_von_diamonds':  FieldDescriptor(0x6e, "Demi von Diamonds",  BoolType()),
        'pilot':              FieldDescriptor(0x6f, "Pilot",              BoolType()),
        'princess_airyn':     FieldDescriptor(0x70, "Princess Airyn",     BoolType()),
        'dirk_yamaoka':       FieldDescriptor(0x71, "Dirk Yamaoka",       BoolType()),
        'guy_spelunky':       FieldDescriptor(0x72, "Guy Spelunky",       BoolType()),
        'classic_spelunky':   FieldDescriptor(0x73, "Classic Spelunky",   BoolType()),
        'terra_tunnel':       FieldDescriptor(0x74, "Terra Tunnel",       BoolType()),
        'hired_hand':         FieldDescriptor(0x75, "Hired Hand",         BoolType()),
        'eggplant_child':     FieldDescriptor(0x76, "Eggplant Child",     BoolType()),
        'shopkeeper':         FieldDescriptor(0x77, "Shopkeeper",         BoolType()),
        'tun':                FieldDescriptor(0x78, "Tun",                BoolType()),
        'yang':               FieldDescriptor(0x79, "Yang",               BoolType()),
        'madame_tusk':        FieldDescriptor(0x7a, "Madame Tusk",        BoolType()),
        'tusk_bodyguard':     FieldDescriptor(0x7b, "Tusk's Bodyguard",   BoolType()),
        'waddler':            FieldDescriptor(0x7c, "Waddler",            BoolType()),
        'caveman_shopkeeper': FieldDescriptor(0x7d, "Caveman Shopkeeper", BoolType()),
        'ghist_shopkeeper':   FieldDescriptor(0x7e, "Ghist Shopkeeper",   BoolType()),
        'van_horsing':        FieldDescriptor(0x7f, "Van Horsing",        BoolType()),
        'parsley':            FieldDescriptor(0x80, "Parsley",            BoolType()),
        'parsnip':            FieldDescriptor(0x81, "Parsnip",            BoolType()),
        'parmesan':           FieldDescriptor(0x82, "Parmesan",           BoolType()),
        'sparrow':            FieldDescriptor(0x83, "Sparrow",            BoolType()),
        'beg':                FieldDescriptor(0x84, "Beg",                BoolType()),
        'eggplant_king':      FieldDescriptor(0x85, "Eggplant King",      BoolType()),
    }),
    'items_discovered': Category('Journal: Items Discovered', {
        'rope_pile':         FieldDescriptor(0x86, "Rope Pile",         BoolType()),
        'bomb_bag':          FieldDescriptor(0x87, "Bomb Bag",          BoolType()),
        'bomb_box':          FieldDescriptor(0x88, "Bomb Box",          BoolType()),
        'paste':             FieldDescriptor(0x89, "Paste",             BoolType()),
        'spectacles':        FieldDescriptor(0x8a, "Spectacles",        BoolType()),
        'climbing_gloves':   FieldDescriptor(0x8b, "Climbing Gloves",   BoolType()),
        'pitcher_mitt':      FieldDescriptor(0x8c, "Pitcher's Mitt",    BoolType()),
        'spring_shoes':      FieldDescriptor(0x8d, "Spring Shoes",      BoolType()),
        'spike_shoes':       FieldDescriptor(0x8e, "Spike Shoes",       BoolType()),
        'compass':           FieldDescriptor(0x8f, "Compass",           BoolType()),
        'alien_compass':     FieldDescriptor(0x90, "Alien Compass",     BoolType()),
        'parachute':         FieldDescriptor(0x91, "Parachute",         BoolType()),
        'udjat_eye':         FieldDescriptor(0x92, "Udjat Eye",         BoolType()),
        'kapala':            FieldDescriptor(0x93, "Kapala",            BoolType()),
        'hedjet':            FieldDescriptor(0x94, "Hedjet",            BoolType()),
        'crown':             FieldDescriptor(0x95, "Crown",             BoolType()),
        'eggplant_crown':    FieldDescriptor(0x96, "Eggplant Crown",    BoolType()),
        'the_true_crown':    FieldDescriptor(0x97, "The True Crown",    BoolType()),
        'ankh':              FieldDescriptor(0x98, "Ankh",              BoolType()),
        'tablet_of_destiny': FieldDescriptor(0x99, "Tablet of Destiny", BoolType()),
        'skeleton_key':      FieldDescriptor(0x9a, "Skeleton Key",      BoolType()),
        'royal_jelly':       FieldDescriptor(0x9b, "Royal Jelly",       BoolType()),
        'cape':              FieldDescriptor(0x9c, "Cape",              BoolType()),
        'vlad_cape':         FieldDescriptor(0x9d, "Vlad's Cape",       BoolType()),
        'jetpack':           FieldDescriptor(0x9e, "Jetpack",           BoolType()),
        'telepack':          FieldDescriptor(0x9f, "Telepack",          BoolType()),
        'hoverpack':         FieldDescriptor(0xa0, "Hoverpack",         BoolType()),
        'powerpack':         FieldDescriptor(0xa1, "Powerpack",         BoolType()),
        'webgun':            FieldDescriptor(0xa2, "Webgun",            BoolType()),
        'shotgun':           FieldDescriptor(0xa3, "Shotgun",           BoolType()),
        'freeze_ray':        FieldDescriptor(0xa4, "Freeze Ray",        BoolType()),
        'clone_gun':         FieldDescriptor(0xa5, "Clone Gun",         BoolType()),
        'crossbow':          FieldDescriptor(0xa6, "Crossbow",          BoolType()),
        'camera':            FieldDescriptor(0xa7, "Camera",            BoolType()),
        'teleporter':        FieldDescriptor(0xa8, "Teleporter",        BoolType()),
        'mattock':           FieldDescriptor(0xa9, "Mattock",           BoolType()),
        'boomerang':         FieldDescriptor(0xaa, "Boomerang",         BoolType()),
        'machete':           FieldDescriptor(0xab, "Machete",           BoolType()),
        'excalibure':        FieldDescriptor(0xac, "Excalibure",        BoolType()),
        'broken_sword':      FieldDescriptor(0xad, "Broken Sword",      BoolType()),
        'plasma_cannon':     FieldDescriptor(0xae, "Plasma Cannon",     BoolType()),
        'scepter':           FieldDescriptor(0xaf, "Scepter",           BoolType()),
        'hou_yi_bow':        FieldDescriptor(0xb0, "Hou Yi's Bow",      BoolType()),
        'arrow_of_light':    FieldDescriptor(0xb1, "Arrow of Light",    BoolType()),
        'wooden_shield':     FieldDescriptor(0xb2, "Wooden Shield",     BoolType()),
        'metal_shield':      FieldDescriptor(0xb3, "Metal Shield",      BoolType()),
        'idol':              FieldDescriptor(0xb4, "Idol",              BoolType()),
        'the_tusk_idol':     FieldDescriptor(0xb5, "The Tusk Idol",     BoolType()),
        'curse_pot':         FieldDescriptor(0xb6, "Curse Pot",         BoolType()),
        'ushabti':           FieldDescriptor(0xb7, "Ushabti",           BoolType()),
        'eggplant':          FieldDescriptor(0xb8, "Eggplant",          BoolType()),
        'cooked_turkey':     FieldDescriptor(0xb9, "Cooked Turkey",     BoolType()),
        'elixir':            FieldDescriptor(0xba, "Elixir",            BoolType()),
        'four_leaf_clover':  FieldDescriptor(0xbb, "Four-Leaf Clover",  BoolType()),
    }),
    'traps_discovered': Category('Journal: Traps Discovered', {
        'spikes':           FieldDescriptor(0xbc, "Spikes",           BoolType()),
        'arrow_trap':       FieldDescriptor(0xbd, "Arrow Trap",       BoolType()),
        'bear_trap':        FieldDescriptor(0xbe, "Bear Trap",        BoolType()),
        'log_trap':         FieldDescriptor(0xbf, "Log Trap",         BoolType()),
        'spear_trap':       FieldDescriptor(0xc0, "Spear Trap",       BoolType()),
        'thorny_vine':      FieldDescriptor(0xc1, "Thorny Vine",      BoolType()),
        'snap_trap':        FieldDescriptor(0xc2, "Snap Trap",        BoolType()),
        'powder_box':       FieldDescriptor(0xc3, "Powder Box",       BoolType()),
        'falling_platform': FieldDescriptor(0xc4, "Falling Platform", BoolType()),
        'spikeball':        FieldDescriptor(0xc5, "Spikeball",        BoolType()),
        'lion_trap':        FieldDescriptor(0xc6, "Lion Trap",        BoolType()),
        'giant_clam':       FieldDescriptor(0xc7, "Giant Clam",       BoolType()),
        'sliding_wall':     FieldDescriptor(0xc8, "Sliding Wall",     BoolType()),
        'crush_trap':       FieldDescriptor(0xc9, "Crush Trap",       BoolType()),
        'giant_crush_trap': FieldDescriptor(0xca, "Giant Crush Trap", BoolType()),
        'boulder':          FieldDescriptor(0xcb, "Boulder",          BoolType()),
        'spring_trap':      FieldDescriptor(0xcc, "Spring Trap",      BoolType()),
        'landmine':         FieldDescriptor(0xcd, "Landmine",         BoolType()),
        'laser_trap':       FieldDescriptor(0xce, "Laser Trap",       BoolType()),
        'spark_trap':       FieldDescriptor(0xcf, "Spark Trap",       BoolType()),
        'frog_trap':        FieldDescriptor(0xd0, "Frog Trap",        BoolType()),
        'sticky_trap':      FieldDescriptor(0xd1, "Sticky Trap",      BoolType()),
        'bone_drop':        FieldDescriptor(0xd2, "Bone Drop",        BoolType()),
        'egg_sac':          FieldDescriptor(0xd3, "Egg Sac",          BoolType()),
    }),
    'character_unlocks': Category('Character Unlocks', {
        'unlock_field': FieldDescriptor(0xe6, "Unlock field", FlagType(4, CharacterUnlockFlags)),
    }),
    'shortcut_unlocks': Category('Shortcut Unlocks', {
        'shortcut_progress': FieldDescriptor(0xeb, "Shortcut Progress", EnumType(1, ShortcutProgressEnum)),
    }),
    'bestiary_defeated': Category('Bestiary Defeated Counts', {
        'snake':             FieldDescriptor(0x0ee, "Snake",             Int32Type()),
        'spider':            FieldDescriptor(0x0f2, "Spider",            Int32Type()),
        'bat':               FieldDescriptor(0x0f6, "Bat",               Int32Type()),
        'caveman':           FieldDescriptor(0x0fa, "Caveman",           Int32Type()),
        'skeleton':          FieldDescriptor(0x0fe, "Skeleton",          Int32Type()),
        'horned_lizard':     FieldDescriptor(0x102, "Horned Lizard",     Int32Type()),
        'moles':             FieldDescriptor(0x106, "Moles",             Int32Type()),
        'quillback':         FieldDescriptor(0x10a, "Quillback",         Int32Type()),
        'mantrap':           FieldDescriptor(0x10e, "Mantrap",           Int32Type()),
        'tiki_man':          FieldDescriptor(0x112, "Tiki Man",          Int32Type()),
        'witch_doctor':      FieldDescriptor(0x116, "Witch Doctor",      Int32Type()),
        'mosquito':          FieldDescriptor(0x11a, "Mosquito",          Int32Type()),
        'monkey':            FieldDescriptor(0x11e, "Monkey",            Int32Type()),
        'hang_spider':       FieldDescriptor(0x122, "Hang Spider",       Int32Type()),
        'giant_spider':      FieldDescriptor(0x126, "Giant Spider",      Int32Type()),
        'magmar':            FieldDescriptor(0x12a, "Magmar",            Int32Type()),
        'robot':             FieldDescriptor(0x12e, "Robot",             Int32Type()),
        'fire_bug':          FieldDescriptor(0x132, "Fire Bug",          Int32Type()),
        'imp':               FieldDescriptor(0x136, "Imp",               Int32Type()),
        'lavamander':        FieldDescriptor(0x13a, "Lavamander",        Int32Type()),
        'vampire':           FieldDescriptor(0x13e, "Vampire",           Int32Type()),
        'vlad':              FieldDescriptor(0x142, "Vlad",              Int32Type()),
        'olmec':             FieldDescriptor(0x146, "Olmec",             Int32Type()),
        'jiangshi':          FieldDescriptor(0x14a, "Jiangshi",          Int32Type()),
        'jiangshi_assassin': FieldDescriptor(0x14e, "Jiangshi Assassin", Int32Type()),
        'flying_fish':       FieldDescriptor(0x152, "Flying Fish",       Int32Type()),
        'octopy':            FieldDescriptor(0x156, "Octopy",            Int32Type()),
        'hermit_crab':       FieldDescriptor(0x15a, "Hermit Crab",       Int32Type()),
        'pangxie':           FieldDescriptor(0x15e, "Pangxie",           Int32Type()),
        'great_humphead':    FieldDescriptor(0x162, "Great Humphead",    Int32Type()),
        'kingu':             FieldDescriptor(0x166, "Kingu",             Int32Type()),
        'crocman':           FieldDescriptor(0x16a, "Crocman",           Int32Type()),
        'cobra':             FieldDescriptor(0x16e, "Cobra",             Int32Type()),
        'mummy':             FieldDescriptor(0x172, "Mummy",             Int32Type()),
        'sorceress':         FieldDescriptor(0x176, "Sorceress",         Int32Type()),
        'cat_mummy':         FieldDescriptor(0x17a, "Cat Mummy",         Int32Type()),
        'necromancer':       FieldDescriptor(0x17e, "Necromancer",       Int32Type()),
        'anubis':            FieldDescriptor(0x182, "Anubis",            Int32Type()),
        'ammit':             FieldDescriptor(0x186, "Ammit",             Int32Type()),
        'apep':              FieldDescriptor(0x18a, "Apep",              Int32Type()),
        'anubis_ii':         FieldDescriptor(0x18e, "Anubis II",         Int32Type()),
        'osiris':            FieldDescriptor(0x192, "Osiris",            Int32Type()),
        'ufo':               FieldDescriptor(0x196, "Ufo",               Int32Type()),
        'alien':             FieldDescriptor(0x19a, "Alien",             Int32Type()),
        'yeti':              FieldDescriptor(0x19e, "Yeti",              Int32Type()),
        'yeti_king':         FieldDescriptor(0x1a2, "Yeti King",         Int32Type()),
        'yeti_queen':        FieldDescriptor(0x1a6, "Yeti Queen",        Int32Type()),
        'lahamu':            FieldDescriptor(0x1aa, "Lahamu",            Int32Type()),
        'proto_shopkeeper':  FieldDescriptor(0x1ae, "Proto Shopkeeper",  Int32Type()),
        'olmite':            FieldDescriptor(0x1b2, "Olmite",            Int32Type()),
        'lamassu':           FieldDescriptor(0x1b6, "Lamassu",           Int32Type()),
        'tiamat':            FieldDescriptor(0x1ba, "Tiamat",            Int32Type()),
        'tadpole':           FieldDescriptor(0x1be, "Tadpole",           Int32Type()),
        'frog':              FieldDescriptor(0x1c2, "Frog",              Int32Type()),
        'fire_frog':         FieldDescriptor(0x1c6, "Fire Frog",         Int32Type()),
        'goliath_frog':      FieldDescriptor(0x1ca, "Goliath Frog",      Int32Type()),
        'grub':              FieldDescriptor(0x1ce, "Grub",              Int32Type()),
        'giant_fly':         FieldDescriptor(0x1d2, "Giant Fly",         Int32Type()),
        'hundun':            FieldDescriptor(0x1d6, "Hundun",            Int32Type()),
        'eggplant_minister': FieldDescriptor(0x1da, "Eggplant Minister", Int32Type()),
        'eggplup':           FieldDescriptor(0x1de, "Eggplup",           Int32Type()),
        'celestial_jelly':   FieldDescriptor(0x1e2, "Celestial Jelly",   Int32Type()),
        'scorpion':          FieldDescriptor(0x1e6, "Scorpion",          Int32Type()),
        'bee':               FieldDescriptor(0x1ea, "Bee",               Int32Type()),
        'queen_bee':         FieldDescriptor(0x1ee, "Queen Bee",         Int32Type()),
        'scarab':            FieldDescriptor(0x1f2, "Scarab",            Int32Type()),
        'golden_monkey':     FieldDescriptor(0x1f6, "Golden Monkey",     Int32Type()),
        'leprechaun':        FieldDescriptor(0x1fa, "Leprechaun",        Int32Type()),
        'monty':             FieldDescriptor(0x1fe, "Monty",             Int32Type()),
        'percy':             FieldDescriptor(0x202, "Percy",             Int32Type()),
        'poochi':            FieldDescriptor(0x206, "Poochi",            Int32Type()),
        'ghist':             FieldDescriptor(0x20a, "Ghist",             Int32Type()),
        'ghost':             FieldDescriptor(0x20e, "Ghost",             Int32Type()),
        'cave_turkey':       FieldDescriptor(0x212, "Cave Turkey",       Int32Type()),
        'rock_dog':          FieldDescriptor(0x216, "Rock Dog",          Int32Type()),
        'axolotl':           FieldDescriptor(0x21a, "Axolotl",           Int32Type()),
        'qilin':             FieldDescriptor(0x21e, "Qilin",             Int32Type()),
        'mech_rider':        FieldDescriptor(0x222, "Mech Rider",        Int32Type()),
    }),
    'bestiary_killed_by': Category('Bestiary Killed by Counts', {
        'snake':             FieldDescriptor(0x226, "Snake",             Int32Type()),
        'spider':            FieldDescriptor(0x22a, "Spider",            Int32Type()),
        'bat':               FieldDescriptor(0x22e, "Bat",               Int32Type()),
        'caveman':           FieldDescriptor(0x232, "Caveman",           Int32Type()),
        'skeleton':          FieldDescriptor(0x236, "Skeleton",          Int32Type()),
        'horned_lizard':     FieldDescriptor(0x23a, "Horned Lizard",     Int32Type()),
        'moles':             FieldDescriptor(0x23e, "Moles",             Int32Type()),
        'quillback':         FieldDescriptor(0x242, "Quillback",         Int32Type()),
        'mantrap':           FieldDescriptor(0x246, "Mantrap",           Int32Type()),
        'tiki_man':          FieldDescriptor(0x24a, "Tiki Man",          Int32Type()),
        'witch_doctor':      FieldDescriptor(0x24e, "Witch Doctor",      Int32Type()),
        'mosquito':          FieldDescriptor(0x252, "Mosquito",          Int32Type()),
        'monkey':            FieldDescriptor(0x256, "Monkey",            Int32Type()),
        'hang_spider':       FieldDescriptor(0x25a, "Hang Spider",       Int32Type()),
        'giant_spider':      FieldDescriptor(0x25e, "Giant Spider",      Int32Type()),
        'magmar':            FieldDescriptor(0x262, "Magmar",            Int32Type()),
        'robot':             FieldDescriptor(0x266, "Robot",             Int32Type()),
        'fire_bug':          FieldDescriptor(0x26a, "Fire Bug",          Int32Type()),
        'imp':               FieldDescriptor(0x26e, "Imp",               Int32Type()),
        'lavamander':        FieldDescriptor(0x272, "Lavamander",        Int32Type()),
        'vampire':           FieldDescriptor(0x276, "Vampire",           Int32Type()),
        'vlad':              FieldDescriptor(0x27a, "Vlad",              Int32Type()),
        'olmec':             FieldDescriptor(0x27e, "Olmec",             Int32Type()),
        'jiangshi':          FieldDescriptor(0x282, "Jiangshi",          Int32Type()),
        'jiangshi_assassin': FieldDescriptor(0x286, "Jiangshi Assassin", Int32Type()),
        'flying_fish':       FieldDescriptor(0x28a, "Flying Fish",       Int32Type()),
        'octopy':            FieldDescriptor(0x28e, "Octopy",            Int32Type()),
        'hermit_crab':       FieldDescriptor(0x292, "Hermit Crab",       Int32Type()),
        'pangxie':           FieldDescriptor(0x296, "Pangxie",           Int32Type()),
        'great_humphead':    FieldDescriptor(0x29a, "Great Humphead",    Int32Type()),
        'kingu':             FieldDescriptor(0x29e, "Kingu",             Int32Type()),
        'crocman':           FieldDescriptor(0x2a2, "Crocman",           Int32Type()),
        'cobra':             FieldDescriptor(0x2a6, "Cobra",             Int32Type()),
        'mummy':             FieldDescriptor(0x2aa, "Mummy",             Int32Type()),
        'sorceress':         FieldDescriptor(0x2ae, "Sorceress",         Int32Type()),
        'cat_mummy':         FieldDescriptor(0x2b2, "Cat Mummy",         Int32Type()),
        'necromancer':       FieldDescriptor(0x2b6, "Necromancer",       Int32Type()),
        'anubis':            FieldDescriptor(0x2ba, "Anubis",            Int32Type()),
        'ammit':             FieldDescriptor(0x2be, "Ammit",             Int32Type()),
        'apep':              FieldDescriptor(0x2c2, "Apep",              Int32Type()),
        'anubis_ii':         FieldDescriptor(0x2c6, "Anubis II",         Int32Type()),
        'osiris':            FieldDescriptor(0x2ca, "Osiris",            Int32Type()),
        'ufo':               FieldDescriptor(0x2ce, "Ufo",               Int32Type()),
        'alien':             FieldDescriptor(0x2d2, "Alien",             Int32Type()),
        'yeti':              FieldDescriptor(0x2d6, "Yeti",              Int32Type()),
        'yeti_king':         FieldDescriptor(0x2da, "Yeti King",         Int32Type()),
        'yeti_queen':        FieldDescriptor(0x2de, "Yeti Queen",        Int32Type()),
        'lahamu':            FieldDescriptor(0x2e2, "Lahamu",            Int32Type()),
        'proto_shopkeeper':  FieldDescriptor(0x2e6, "Proto Shopkeeper",  Int32Type()),
        'olmite':            FieldDescriptor(0x2ea, "Olmite",            Int32Type()),
        'lamassu':           FieldDescriptor(0x2ee, "Lamassu",           Int32Type()),
        'tiamat':            FieldDescriptor(0x2f2, "Tiamat",            Int32Type()),
        'tadpole':           FieldDescriptor(0x2f6, "Tadpole",           Int32Type()),
        'frog':              FieldDescriptor(0x2fa, "Frog",              Int32Type()),
        'fire_frog':         FieldDescriptor(0x2fe, "Fire Frog",         Int32Type()),
        'goliath_frog':      FieldDescriptor(0x302, "Goliath Frog",      Int32Type()),
        'grub':              FieldDescriptor(0x306, "Grub",              Int32Type()),
        'giant_fly':         FieldDescriptor(0x30a, "Giant Fly",         Int32Type()),
        'hundun':            FieldDescriptor(0x30e, "Hundun",            Int32Type()),
        'eggplant_minister': FieldDescriptor(0x312, "Eggplant Minister", Int32Type()),
        'eggplup':           FieldDescriptor(0x316, "Eggplup",           Int32Type()),
        'celestial_jelly':   FieldDescriptor(0x31a, "Celestial Jelly",   Int32Type()),
        'scorpion':          FieldDescriptor(0x31e, "Scorpion",          Int32Type()),
        'bee':               FieldDescriptor(0x322, "Bee",               Int32Type()),
        'queen_bee':         FieldDescriptor(0x326, "Queen Bee",         Int32Type()),
        'scarab':            FieldDescriptor(0x32a, "Scarab",            Int32Type()),
        'golden_monkey':     FieldDescriptor(0x32e, "Golden Monkey",     Int32Type()),
        'leprechaun':        FieldDescriptor(0x332, "Leprechaun",        Int32Type()),
        'monty':             FieldDescriptor(0x336, "Monty",             Int32Type()),
        'percy':             FieldDescriptor(0x33a, "Percy",             Int32Type()),
        'poochi':            FieldDescriptor(0x33e, "Poochi",            Int32Type()),
        'ghist':             FieldDescriptor(0x342, "Ghist",             Int32Type()),
        'ghost':             FieldDescriptor(0x346, "Ghost",             Int32Type()),
        'cave_turkey':       FieldDescriptor(0x34a, "Cave Turkey",       Int32Type()),
        'rock_dog':          FieldDescriptor(0x34e, "Rock Dog",          Int32Type()),
        'axolotl':           FieldDescriptor(0x352, "Axolotl",           Int32Type()),
        'qilin':             FieldDescriptor(0x356, "Qilin",             Int32Type()),
        'mech_rider':        FieldDescriptor(0x35a, "Mech Rider",        Int32Type()),
    }),
    'people_defeated': Category('People Defeated Counts', {
        'hired_hand':         FieldDescriptor(0x3b2, "Hired Hand",         Int32Type()),
        'eggplant_child':     FieldDescriptor(0x3b6, "Eggplant Child",     Int32Type()),
        'shopkeeper':         FieldDescriptor(0x3ba, "Shopkeeper",         Int32Type()),
        'tun':                FieldDescriptor(0x3be, "Tun",                Int32Type()),
        'yang':               FieldDescriptor(0x3c2, "Yang",               Int32Type()),
        'madame_tusk':        FieldDescriptor(0x3c6, "Madame Tusk",        Int32Type()),
        'tusk_bodyguard':     FieldDescriptor(0x3ca, "Tusk's Bodyguard",   Int32Type()),
        'waddler':            FieldDescriptor(0x3ce, "Waddler",            Int32Type()),
        'caveman_shopkeeper': FieldDescriptor(0x3d2, "Caveman Shopkeeper", Int32Type()),
        'ghist_shopkeeper':   FieldDescriptor(0x3d6, "Ghist Shopkeeper",   Int32Type()),
        'van_horsing':        FieldDescriptor(0x3da, "Van Horsing",        Int32Type()),
        'parsley':            FieldDescriptor(0x3de, "Parsley",            Int32Type()),
        'parsnip':            FieldDescriptor(0x3e2, "Parsnip",            Int32Type()),
        'parmesan':           FieldDescriptor(0x3e6, "Parmesan",           Int32Type()),
        'sparrow':            FieldDescriptor(0x3ea, "Sparrow",            Int32Type()),
        'beg':                FieldDescriptor(0x3ee, "Beg",                Int32Type()),
        'eggplant_king':      FieldDescriptor(0x3f2, "Eggplant King",      Int32Type()),
    }),
    'people_killed_by': Category('People Killed by Counts', {
        'hired_hand':         FieldDescriptor(0x44a, "Hired Hand",         Int32Type()),
        'eggplant_child':     FieldDescriptor(0x44e, "Eggplant Child",     Int32Type()),
        'shopkeeper':         FieldDescriptor(0x452, "Shopkeeper",         Int32Type()),
        'tun':                FieldDescriptor(0x456, "Tun",                Int32Type()),
        'yang':               FieldDescriptor(0x45a, "Yang",               Int32Type()),
        'madame_tusk':        FieldDescriptor(0x45e, "Madame Tusk",        Int32Type()),
        'tusk_bodyguard':     FieldDescriptor(0x462, "Tusk's Bodyguard",   Int32Type()),
        'waddler':            FieldDescriptor(0x466, "Waddler",            Int32Type()),
        'caveman_shopkeeper': FieldDescriptor(0x46a, "Caveman Shopkeeper", Int32Type()),
        'ghist_shopkeeper':   FieldDescriptor(0x46e, "Ghist Shopkeeper",   Int32Type()),
        'van_horsing':        FieldDescriptor(0x472, "Van Horsing",        Int32Type()),
        'parsley':            FieldDescriptor(0x476, "Parsley",            Int32Type()),
        'parsnip':            FieldDescriptor(0x47a, "Parsnip",            Int32Type()),
        'parmesan':           FieldDescriptor(0x47e, "Parmesan",           Int32Type()),
        'sparrow':            FieldDescriptor(0x482, "Sparrow",            Int32Type()),
        'beg':                FieldDescriptor(0x486, "Beg",                Int32Type()),
        'eggplant_king':      FieldDescriptor(0x48a, "Eggplant King",      Int32Type()),
    }),
    'player_profile': Category('Player Profile', {
        'plays':         FieldDescriptor(0x048e, "Plays",         Int32Type()),
        'deaths':        FieldDescriptor(0x0492, "Deaths",        Int32Type()),
        'normal_wins':   FieldDescriptor(0x0496, "Normal Wins",   Int32Type()),
        'hard_wins':     FieldDescriptor(0x049a, "Hard Wins",     Int32Type()),
        'special_wins':  FieldDescriptor(0x049e, "Special Wins",  Int32Type()),
        'sum_of_score':  FieldDescriptor(0x04a2, "Sum of Score",  Int64Type()),
        'top_score':     FieldDescriptor(0x04aa, "Top Score",     Int32Type()),
        'deepest_area':  FieldDescriptor(0x04ae, "Deepest Area",  ByteType()),
        'deepest_level': FieldDescriptor(0x04af, "Deepest Level", ByteType()),
        'sum_of_time':   FieldDescriptor(0x288e, "Sum of Time",   Int64Type()),
        'best_time':     FieldDescriptor(0x2896, "Best Time",     Int32Type()),
    }),
    'area_deaths': Category('Area Death Counts', {
        'level_1_1':  FieldDescriptor(0x08b2, "1-1",  Int32Type()),
        'level_1_2':  FieldDescriptor(0x08b6, "1-2",  Int32Type()),
        'level_1_3':  FieldDescriptor(0x08ba, "1-3",  Int32Type()),
        'level_1_4':  FieldDescriptor(0x08be, "1-4",  Int32Type()),
        'level_2_1':  FieldDescriptor(0x0cae, "2-1",  Int32Type()),
        'level_2_2':  FieldDescriptor(0x0cb2, "2-2",  Int32Type()),
        'level_2_3':  FieldDescriptor(0x0cb6, "2-3",  Int32Type()),
        'level_2_4':  FieldDescriptor(0x0cba, "2-4",  Int32Type()),
        'level_3_1':  FieldDescriptor(0x10aa, "3-1",  Int32Type()),
        'level_4_1':  FieldDescriptor(0x14a6, "4-1",  Int32Type()),
        'level_4_2':  FieldDescriptor(0x14aa, "4-2",  Int32Type()),
        'level_4_3':  FieldDescriptor(0x14ae, "4-3",  Int32Type()),
        'level_4_4':  FieldDescriptor(0x14b2, "4-4",  Int32Type()),
        'level_5_1':  FieldDescriptor(0x18a2, "5-1",  Int32Type()),
        'level_6_1':  FieldDescriptor(0x1c9e, "6-1",  Int32Type()),
        'level_6_2':  FieldDescriptor(0x1ca2, "6-2",  Int32Type()),
        'level_6_3':  FieldDescriptor(0x1ca6, "6-3",  Int32Type()),
        'level_6_4':  FieldDescriptor(0x1caa, "6-4",  Int32Type()),
        'level_7_1':  FieldDescriptor(0x209a, "7-1",  Int32Type()),
        'level_7_2':  FieldDescriptor(0x209e, "7-2",  Int32Type()),
        'level_7_3':  FieldDescriptor(0x20a2, "7-3",  Int32Type()),
        'level_7_4':  FieldDescriptor(0x20a6, "7-4",  Int32Type()),
        'level_7_5':  FieldDescriptor(0x20aa, "7-5",  Int32Type()),
        'level_7_6':  FieldDescriptor(0x20ae, "7-6",  Int32Type()),
        'level_7_7':  FieldDescriptor(0x20b2, "7-7",  Int32Type()),
        'level_7_8':  FieldDescriptor(0x20b6, "7-8",  Int32Type()),
        'level_7_9':  FieldDescriptor(0x20ba, "7-9",  Int32Type()),
        'level_7_10': FieldDescriptor(0x20be, "7-10", Int32Type()),
        'level_7_11': FieldDescriptor(0x20c2, "7-11", Int32Type()),
        'level_7_12': FieldDescriptor(0x20c6, "7-12", Int32Type()),
        'level_7_13': FieldDescriptor(0x20ca, "7-13", Int32Type()),
        'level_7_14': FieldDescriptor(0x20ce, "7-14", Int32Type()),
        'level_7_15': FieldDescriptor(0x20d2, "7-15", Int32Type()),
        'level_7_16': FieldDescriptor(0x20d6, "7-16", Int32Type()),
        'level_7_17': FieldDescriptor(0x20da, "7-17", Int32Type()),
        'level_7_18': FieldDescriptor(0x20de, "7-18", Int32Type()),
        'level_7_19': FieldDescriptor(0x20e2, "7-19", Int32Type()),
        'level_7_20': FieldDescriptor(0x20e6, "7-20", Int32Type()),
        'level_7_21': FieldDescriptor(0x20ea, "7-21", Int32Type()),
        'level_7_22': FieldDescriptor(0x20ee, "7-22", Int32Type()),
        'level_7_23': FieldDescriptor(0x20f2, "7-23", Int32Type()),
        'level_7_24': FieldDescriptor(0x20f6, "7-24", Int32Type()),
        'level_7_25': FieldDescriptor(0x20fa, "7-25", Int32Type()),
        'level_7_26': FieldDescriptor(0x20fe, "7-26", Int32Type()),
        'level_7_27': FieldDescriptor(0x2102, "7-27", Int32Type()),
        'level_7_28': FieldDescriptor(0x2106, "7-28", Int32Type()),
        'level_7_29': FieldDescriptor(0x210a, "7-29", Int32Type()),
        'level_7_30': FieldDescriptor(0x210e, "7-30", Int32Type()),
        'level_7_31': FieldDescriptor(0x2112, "7-31", Int32Type()),
        'level_7_32': FieldDescriptor(0x2116, "7-32", Int32Type()),
        'level_7_33': FieldDescriptor(0x211a, "7-33", Int32Type()),
        'level_7_34': FieldDescriptor(0x211e, "7-34", Int32Type()),
        'level_7_35': FieldDescriptor(0x2122, "7-35", Int32Type()),
        'level_7_36': FieldDescriptor(0x2126, "7-36", Int32Type()),
        'level_7_37': FieldDescriptor(0x212a, "7-37", Int32Type()),
        'level_7_38': FieldDescriptor(0x212e, "7-38", Int32Type()),
        'level_7_39': FieldDescriptor(0x2132, "7-39", Int32Type()),
        'level_7_40': FieldDescriptor(0x2136, "7-40", Int32Type()),
        'level_7_41': FieldDescriptor(0x213a, "7-41", Int32Type()),
        'level_7_42': FieldDescriptor(0x213e, "7-42", Int32Type()),
        'level_7_43': FieldDescriptor(0x2142, "7-43", Int32Type()),
        'level_7_44': FieldDescriptor(0x2146, "7-44", Int32Type()),
        'level_7_45': FieldDescriptor(0x214a, "7-45", Int32Type()),
        'level_7_46': FieldDescriptor(0x214e, "7-46", Int32Type()),
        'level_7_47': FieldDescriptor(0x2152, "7-47", Int32Type()),
        'level_7_48': FieldDescriptor(0x2156, "7-48", Int32Type()),
        'level_7_49': FieldDescriptor(0x215a, "7-49", Int32Type()),
        'level_7_50': FieldDescriptor(0x215e, "7-50", Int32Type()),
        'level_7_51': FieldDescriptor(0x2162, "7-51", Int32Type()),
        'level_7_52': FieldDescriptor(0x2166, "7-52", Int32Type()),
        'level_7_53': FieldDescriptor(0x216a, "7-53", Int32Type()),
        'level_7_54': FieldDescriptor(0x216e, "7-54", Int32Type()),
        'level_7_55': FieldDescriptor(0x2172, "7-55", Int32Type()),
        'level_7_56': FieldDescriptor(0x2176, "7-56", Int32Type()),
        'level_7_57': FieldDescriptor(0x217a, "7-57", Int32Type()),
        'level_7_58': FieldDescriptor(0x217e, "7-58", Int32Type()),
        'level_7_59': FieldDescriptor(0x2182, "7-59", Int32Type()),
        'level_7_60': FieldDescriptor(0x2186, "7-60", Int32Type()),
        'level_7_61': FieldDescriptor(0x218a, "7-61", Int32Type()),
        'level_7_62': FieldDescriptor(0x218e, "7-62", Int32Type()),
        'level_7_63': FieldDescriptor(0x2192, "7-63", Int32Type()),
        'level_7_64': FieldDescriptor(0x2196, "7-64", Int32Type()),
        'level_7_65': FieldDescriptor(0x219a, "7-65", Int32Type()),
        'level_7_66': FieldDescriptor(0x219e, "7-66", Int32Type()),
        'level_7_67': FieldDescriptor(0x21a2, "7-67", Int32Type()),
        'level_7_68': FieldDescriptor(0x21a6, "7-68", Int32Type()),
        'level_7_69': FieldDescriptor(0x21aa, "7-69", Int32Type()),
        'level_7_70': FieldDescriptor(0x21ae, "7-70", Int32Type()),
        'level_7_71': FieldDescriptor(0x21b2, "7-71", Int32Type()),
        'level_7_72': FieldDescriptor(0x21b6, "7-72", Int32Type()),
        'level_7_73': FieldDescriptor(0x21ba, "7-73", Int32Type()),
        'level_7_74': FieldDescriptor(0x21be, "7-74", Int32Type()),
        'level_7_75': FieldDescriptor(0x21c2, "7-75", Int32Type()),
        'level_7_76': FieldDescriptor(0x21c6, "7-76", Int32Type()),
        'level_7_77': FieldDescriptor(0x21ca, "7-77", Int32Type()),
        'level_7_78': FieldDescriptor(0x21ce, "7-78", Int32Type()),
        'level_7_79': FieldDescriptor(0x21d2, "7-79", Int32Type()),
        'level_7_80': FieldDescriptor(0x21d6, "7-80", Int32Type()),
        'level_7_81': FieldDescriptor(0x21da, "7-81", Int32Type()),
        'level_7_82': FieldDescriptor(0x21de, "7-82", Int32Type()),
        'level_7_83': FieldDescriptor(0x21e2, "7-83", Int32Type()),
        'level_7_84': FieldDescriptor(0x21e6, "7-84", Int32Type()),
        'level_7_85': FieldDescriptor(0x21ea, "7-85", Int32Type()),
        'level_7_86': FieldDescriptor(0x21ee, "7-86", Int32Type()),
        'level_7_87': FieldDescriptor(0x21f2, "7-87", Int32Type()),
        'level_7_88': FieldDescriptor(0x21f6, "7-88", Int32Type()),
        'level_7_89': FieldDescriptor(0x21fa, "7-89", Int32Type()),
        'level_7_90': FieldDescriptor(0x21fe, "7-90", Int32Type()),
        'level_7_91': FieldDescriptor(0x2202, "7-91", Int32Type()),
        'level_7_92': FieldDescriptor(0x2206, "7-92", Int32Type()),
        'level_7_93': FieldDescriptor(0x220a, "7-93", Int32Type()),
        'level_7_94': FieldDescriptor(0x220e, "7-94", Int32Type()),
        'level_7_95': FieldDescriptor(0x2212, "7-95", Int32Type()),
        'level_7_96': FieldDescriptor(0x2216, "7-96", Int32Type()),
        'level_7_97': FieldDescriptor(0x221a, "7-97", Int32Type()),
        'level_7_98': FieldDescriptor(0x221e, "7-98", Int32Type()),
    }),
    'character_deaths': Category('Character Death Counts', {
        'ana_spelunky':      FieldDescriptor(0x289a, "Ana Spelunky",      Int32Type()),
        'margaret_tunnel':   FieldDescriptor(0x289e, "Margaret Tunnel",   Int32Type()),
        'colin_northward':   FieldDescriptor(0x28a2, "Colin Northward",   Int32Type()),
        'roffy_sloth':       FieldDescriptor(0x28a6, "Roffy D. Sloth",    Int32Type()),
        'alto_singh':        FieldDescriptor(0x28aa, "Alto Singh",        Int32Type()),
        'liz_mutton':        FieldDescriptor(0x28ae, "Liz Mutton",        Int32Type()),
        'nekka_the_eagle':   FieldDescriptor(0x28b2, "Nekka the Eagle",   Int32Type()),
        'lise_project':      FieldDescriptor(0x28b6, "Lise Project",      Int32Type()),
        'coco_von_diamonds': FieldDescriptor(0x28ba, "Coco von Diamonds", Int32Type()),
        'manfred_tunnel':    FieldDescriptor(0x28be, "Manfred Tunnel",    Int32Type()),
        'little_jay':        FieldDescriptor(0x28c2, "Little Jay",        Int32Type()),
        'tina_flan':         FieldDescriptor(0x28c6, "Tina Flan",         Int32Type()),
        'valerie_crump':     FieldDescriptor(0x28ca, "Valerie Crump",     Int32Type()),
        'au':                FieldDescriptor(0x28ce, "Au",                Int32Type()),
        'demi_von_diamonds': FieldDescriptor(0x28d2, "Demi von Diamonds", Int32Type()),
        'pilot':             FieldDescriptor(0x28d6, "Pilot",             Int32Type()),
        'princess_airyn':    FieldDescriptor(0x28da, "Princess Airyn",    Int32Type()),
        'dirk_yamaoka':      FieldDescriptor(0x28de, "Dirk Yamaoka",      Int32Type()),
        'guy_spelunky':      FieldDescriptor(0x28e2, "Guy Spelunky",      Int32Type()),
        'classic_spelunky':  FieldDescriptor(0x28e6, "Classic Spelunky",  Int32Type()),
    }),
    'completion': Category('Completion', {
        'normal':  FieldDescriptor(0x2901, "Completed the game (Normal)",                    BoolType()),
        'ironman': FieldDescriptor(0x2902, "Completed the game with no shortcuts (Ironman)", BoolType()),
        'hard':    FieldDescriptor(0x2903, "Completed the game (Hard)",                      BoolType()),
    }),
    'viewed_player_profile': Category('Viewed Player Profile', {
        'viewed': FieldDescriptor(0x2904, 'Viewed', BoolType()),
    }),
    'unlocked_seeded_runs': Category('Seeded Runs Unlocked', {
        'unlocked': FieldDescriptor(0x2905, 'Unlocked', BoolType()),
    }),
    'last_game_played': Category('Last Game Played', {
        'area':  FieldDescriptor(0x2906, 'Area',  ByteType()),
        'level': FieldDescriptor(0x2907, 'Level', ByteType()),
        'money': FieldDescriptor(0x290a, 'Money', Int32Type()),
        'time':  FieldDescriptor(0x290e, 'Time',  Int32Type()),
    }),
    'character_selected': Category('Player - Character Selection', {
        'player_1': FieldDescriptor(0x2a7a, "Player 1", EnumType(1, CharacterSelectedEnum)),
        'player_2': FieldDescriptor(0x2a7b, "Player 2", EnumType(1, CharacterSelectedEnum)),
        'player_3': FieldDescriptor(0x2a7c, "Player 3", EnumType(1, CharacterSelectedEnum)),
        'player_4': FieldDescriptor(0x2a7d, "Player 4", EnumType(1, CharacterSelectedEnum)),
    }),
    'crc': Category('CRC32', {
        'crc': FieldDescriptor(0x359a, 'CRC32', UInt32Type())
    }),
}



from dataclasses import dataclass

@dataclass
class Entry:
    id_: int
    name: str
    description: str


ENTRIES = {
    "places": [
        Entry(
            1, "Dwelling", "The air here is warm and dry. It smells of campfire smoke and bat guano.",
        ),
        Entry(
            2, "Jungle", "Despite all odds, this dense, tropical forest appears to be teeming with life.",
        ),
        Entry(
            3, "Volcana", "It's as hot as Hell in this lava-filled world! Who knew the Moon could have such a place inside it?",
        ),
        Entry(
            4, "Olmec's Lair", "A shrine dedicated to the mighty Olmec. There is a timeless quality to this place of worship.",
        ),
        Entry(
            5, "Tide Pool", "The ruins of an ancient empire lay here at the edge of a vast, underground ocean.",
        ),
        Entry(
            6, "Abzu", "This palatial sanctuary was built over the deepest part of the sea.",
        ),
        Entry(
            7, "Temple of Anubis", "A massive tomb build by powerful pharaohs. On its walls are carved the stories of the past, present, and future.",
        ),
        Entry(
            8, "The City of Gold", "A legendary place to all treasure hunters! Every brick is made from solid gold.",
        ),
        Entry(
            9, "Duat", "A desert of ash and fire, burning beneath an eternal blood moon. The domain of the god Osiris.",
        ),
        Entry(
            10, "Ice Caves", "Life still thrives within the frozen core of this giant meteorite.",
        ),
        Entry(
            11, "Neo Babylon", "The capital city of the Olmites, a place of high technology, untold wonders, and endless stimulation.",
        ),
        Entry(
            12, "Tiamat's Throne", "Where Queen Tiamat gave birth to the World. Now a seat of power from which she guides her unruly children.",
        ),
        Entry(
            13, "Sunken City", "In the bottom of the Well, the First City sleeps.",
        ),
        Entry(
            14, "Eggplant World", "Once the palace of a mighty tyrant, now overgrown with peaceful eggplants.",
        ),
        Entry(
            15, "Hundun's Hideaway", "At the lowest point of the Well is where the Great Dreamer muddles about, always felt but never known.",
        ),
        Entry(
            16, "Cosmic Ocean", "The roiling waters of creation that flow from Hundun's childish weeping. Here one may find the beginning and the end.",
        ),
    ],
    "people": [
        Entry(
            1, "Ana Spelunky", "She's come to the Moon to find her mom and dad. An easygoing young lady with the heart of a hero.",
        ),
        Entry(
            2, "Margaret Tunnel", "A fascination with the ocean led her away from the family business. Her dream is to buy her own pirate ship.",
        ),
        Entry(
            3, "Colin Northward", "An inquisitive artist and engineer who's obsessed with Olmec and the Curse of the Caves.",
        ),
        Entry(
            4, "Roffy D. Sloth", "He has a lot of deep thoughts, but not a lot that he wants to say out loud.",
        ),
        Entry(
            5, "Alto Singh", "He works in physical education and wants to make sure everyone's taking good care of their health and staying positive.",
        ),
        Entry(
            6, "Liz Mutton", "A tough-as-nails adventurer who is hard to warm up to and needs her solitude, but also cares deeply for her friends.",
        ),
        Entry(
            7, "Nekka the Eagle", "A ferocious warrior chief who wants to confront the gods and question them about the problems of the universe.",
        ),
        Entry(
            8, "LISE Project", "A robot designed to explore and enjoy herself. After 1000 years she'll return to her creators with the data she's collected.",
        ),
        Entry(
            9, "Coco Von Diamonds", "The creator of the world-famous Von Diamonds fashion and lifestyle brand. She's on vacation right now.",
        ),
        Entry(
            10, "Manfred Tunnel", "He's decided to finally try his hand at spelunking after many years of building shortcuts.",
        ),
        Entry(
            11, "Little Jay", "A self-proclaimed \"Spelunkyhead\" trying to become a famous explorer. Gets a little overly excited around his heroes.",
        ),
        Entry(
            12, "Tina Flan", "As a relatively unknown caver she managed to hit King Yama with an eggplant. She tries to downplay her fame!",
        ),
        Entry(
            13, "Valerie Crump", "An avid fan of music and video games. She's trying to find creative inspiration through travel.",
        ),
        Entry(
            14, "Au", "A monk with bright golden skin. By challenging his mind and body he's trying to find peace in a chaotic universe.",
        ),
        Entry(
            15, "Demi Von Diamonds", "The heiress to the Von Diamonds empire. She's lived a pampered life, but is eager to learn more about the world.",
        ),
        Entry(
            16, "Pilot", "A strange little spaceman from a far away planet. He can often be found muttering about his UFO, the Campanella.",
        ),
        Entry(
            17, "Princess Airyn", "A culinary expert, she travels the world looking for the most exquisite tastes to bring home to her kingdom of Heliotropia.",
        ),
        Entry(
            18, "Dirk Yamaoka", "A master ninja always looking to hone his skills. When the Moon called, he answered without hesitation.",
        ),
        Entry(
            19, "Guy Spelunky", "He's spent his life in pursuit of adventure (and treasure). A young man no longer, he's re-evaluating what's most important to him.",
        ),
        Entry(
            20, "Classic Spelunky", "A version of Guy Spelunky from a parallel universe made of large, flat blocks of colors.",
        ),
        Entry(
            21, "Terra Tunnel", "She met Manfred many years ago, when she rescued him from a cave-in. Together they've built a thriving shortcut-building business!",
        ),
        Entry(
            22, "Hired Hand", "He's been trapped in the caves for a very long time and is eager to be free once again.",
        ),
        Entry(
            23, "Eggplant Child", "Separated from their mother, they miss her dearly. The embodiment of curiosity and creative potential.",
        ),
        Entry(
            24, "Shopkeeper", "A card-carrying member of the Shopkeeper's Association. He HATES shoplifting above all else!",
        ),
        Entry(
            25, "Tun", "A mysterious young woman who keeps a close eye on what's happening in the Caves. Beg's sister.",
        ),
        Entry(
            26, "Yang", "The youngest of Yang's 12 sons, who are all named Yang. He loves turkeys.",
        ),
        Entry(
            27, "Madame Tusk", "The most famous resident of the caves, or at least she'd like to think so!",
        ),
        Entry(
            28, "Tusk's Bodyguard", "Madame Tusk pays them extremely well.",
        ),
        Entry(
            29, "Waddler", "He runs an item storage and transportation business. Customers have complained of getting their items back sticky.",
        ),
        Entry(
            30, "Caveman Shopkeeper", "An enterprising caveman trying to acquire more shiny things to show off to his friends.",
        ),
        Entry(
            31, "Ghist Shopkeeper", "In the darkest corners they peddle their wares, hoping to preserve some small shred of their former life.",
        ),
        Entry(
            32, "Van Horsing", "A veteran from the first adventure. He seeks revenge on Vlad for reasons unknown.",
        ),
        Entry(
            33, "Parsley", "The youngest of three sisters. She's cheery but has a quick temper.",
        ),
        Entry(
            34, "Parsnip", "The middle child of three sisters. She loves parties and cats.",
        ),
        Entry(
            35, "Parmesan", "The eldest of three sisters. She's shy and nurturing.",
        ),
        Entry(
            36, "Sparrow", "Hailing from a prominent ninja family, she defied her parents and turned to thievery to challenge authority.",
        ),
        Entry(
            37, "Beg", "In his insecurity he was drawn to Hundun and embraced Chaos. Tun's brother.",
        ),
        Entry(
            38, "Eggplant King", "He was a mighty king... until a brave explorer humbled him. Some might say he's even more powerful now.",
        ),
    ],
    "items": [
        Entry(
            1, "Rope Pile", "3 ropes to help you get to those hard-to-reach places.",
        ),
        Entry(
            2, "Bomb Bag", "A large sack filled with 3 bombs.",
        ),
        Entry(
            3, "Bomb Box", "A flame-resistant box of 12 bombs. The gift every adventurer wants to receive!",
        ),
        Entry(
            4, "Paste", "Made from spider's web. Lather it on your bombs to make them sticky!",
        ),
        Entry(
            5, "Spectacles", "Improves your vision and also makes you look slightly more intelligent.",
        ),
        Entry(
            6, "Climbing Gloves", "The grips of these gloves have small bristles all over them. Just lean into walls to cling to them!",
        ),
        Entry(
            7, "Pitcher's Mitt", "This classic American baseball glove makes you feel like you can throw like a pro.",
        ),
        Entry(
            8, "Spring Shoes", "A sturdy pair of boots equipped with alien spring technology.",
        ),
        Entry(
            9, "Spike Shoes", "A sturdy pair of boots with sharp spikes fitted to the bottom.",
        ),
        Entry(
            10, "Compass", "This classic explorer's tool will point you right to the exit. How convenient!",
        ),
        Entry(
            11, "Alien Compass", "A special compass designed by aliens to help them return home through the inky darkness of space.",
        ),
        Entry(
            12, "Parachute", "It will save your life when you least expect it!",
        ),
        Entry(
            13, "Udjat Eye", "The udjat eye is a symbol of great power. It represents the gods' watchful gaze.",
        ),
        Entry(
            14, "Kapala", "A skull cup bestowed upon Kali's most loyal devotees. Drinking blood from it can improve one's health and enlightenment.",
        ),
        Entry(
            15, "Hedjet", "A white pharaoh's crown, indicating the wearer's status as an envoy for the gods. It glows brilliantly.",
        ),
        Entry(
            16, "Crown", "This royal headgear exudes an aura of unquestionable authority. It's much heavier than it looks!",
        ),
        Entry(
            17, "Eggplant Crown", "The Eggplant King bequeathed his crown to me. I feel at peace with my place in the universe.",
        ),
        Entry(
            18, "The True Crown", "Also known as the cap and bells. A fitting crown for the followers of Hundun.",
        ),
        Entry(
            19, "Ankh", "This coveted amulet was created by the gods for followers deemed worthy enough to have their life extended.",
        ),
        Entry(
            20, "Tablet of Destiny", "A clay tablet on which many esoteric prophecies are written.",
        ),
        Entry(
            21, "Skeleton Key", "The tool of thieves. It opens any lock without breaking.",
        ),
        Entry(
            22, "Royal Jelly", "A delicacy that's highly sought after for its refined taste and healing properties.",
        ),
        Entry(
            23, "Cape", "This dashing fashion accessory can also be used to reduce one's speed while falling.",
        ),
        Entry(
            24, "Vlad's Cape", "The blood-red cape of Vlad the Impaler. It offers impressive mobility.",
        ),
        Entry(
            25, "Jetpack", "It's expensive, but many explorers swear by it. Just keep an eye on your fuel levels!",
        ),
        Entry(
            26, "Telepack", "Allows you to teleport while freeing up your hands! Requires your feet to be off the ground, however.",
        ),
        Entry(
            27, "Hoverpack", "A refinement of Olmite technology. It allows you to hover indefinitely.",
        ),
        Entry(
            28, "Powerpack", "Developed for military use, the power pack imbues its wearer with improved weapon handling abilities.",
        ),
        Entry(
            29, "Webgun", "It shoots synthetic webs that are so close to the real thing that even spiders can't tell the difference",
        ),
        Entry(
            30, "Shotgun", "A basic 12-gauge shotgun. Favored by shopkeepers for self-defense and hunting down shoplifters.",
        ),
        Entry(
            31, "Freeze Ray", "The say the mechanism behind its frozen blast was inspired by the sneezes of the now-extinct mammoth.",
        ),
        Entry(
            32, "Clone Gun", "A one-of-a-kind prototype! The technology required to build one is beyond any known civilization.",
        ),
        Entry(
            33, "Crossbow", "A ranged weapon favored by hunters. It's capable of firing arrows with incredible accuracy.",
        ),
        Entry(
            34, "Camera", "This fancy camera comes with the most powerful flash on the market right now.",
        ),
        Entry(
            35, "Teleporter", "A handheld teleportation device. It's still in beta after all these years!",
        ),
        Entry(
            36, "Mattock", "A useful digging tool that breaks easily. The Tunnel family prefers shovels.",
        ),
        Entry(
            37, "Boomerang", "A curved wooden weapon that will return to its owner after being thrown",
        ),
        Entry(
            38, "Machete", "Its long, broad blade is great for defense or chopping vegetables for dinner.",
        ),
        Entry(
            39, "Excalibur", "The sword of legend. Wielded by kings and queens in times of great calamity.",
        ),
        Entry(
            40, "Broken Sword", "Although perhaps it no longer contains its full strength, it's still a formidable weapon.",
        ),
        Entry(
            41, "Plasma Cannon", "It was built by aliens, but sadly the recoil proved to be too strong for them to use it easily.",
        ),
        Entry(
            42, "Scepter", "A magical tool used by the god Anubis to cleanse souls on their way to judgement.",
        ),
        Entry(
            43, "Hou Yi's Bow", "With this simple but refined bow, Hou Yi shot down the nine suns that scorched the Earth.",
        ),
        Entry(
            44, "Arrow of Light", "A ray of sun reflecting off the Moon's tears.",
        ),
        Entry(
            45, "Wooden Shield", "A flimsy wooden shield carried by Tikimen for some basic defense against the harsh Jungle environment.",
        ),
        Entry(
            46, "Metal Shield", "A highly durable knight's shield made of tempered steel.",
        ),
        Entry(
            47, "Idol", "This valuable golden idol can be brought to the exit or to shops in exchange for gold.",
        ),
        Entry(
            48, "The Tusk Idol", "A walrus idol carved from an extra-large diamond. One of Madame's most precious treasures.",
        ),
        Entry(
            49, "Curse Pot", "A clay pot filled with a toxic substance rumored to be a mixture of ectoplasm and cursed blood.",
        ),
        Entry(
            50, "Ushabti", "A funerary figurine created to accompany someone on their journey to the underworld.",
        ),
        Entry(
            51, "Eggplant", "It is fragile yet full of vigor, reminding us to treasure every one of life's moments.",
        ),
        Entry(
            52, "Cooked Turkey", "A turkey that has been cooked just right. The meat is tender and flavorful.",
        ),
        Entry(
            53, "Elixir", "A restorative potion said to cure any affliction. Royal jelly is mixed in to reduce the bitterness.",
        ),
        Entry(
            54, "Four-Leaf Clover", "A rare clover. It's prized by leprechauns for protection and good fortune.",
        ),
    ],
    "bestiary": [
        Entry(
            1, "Snake", "Be careful! It's liable to throw a hissy fit!",
        ),
        Entry(
            2, "Spider", "Much larger than the household variety. Don't let it get the drop on you!",
        ),
        Entry(
            3, "Bat", "A common cave resident that nonetheless thinks very highly of itself due to its cute appearance.",
        ),
        Entry(
            4, "Caveman", "One of our primitive ancestors. His simple life is somewhat enviable.",
        ),
        Entry(
            5, "Skeleton", "We all have one inside us... and yet they're still so spooky!",
        ),
        Entry(
            6, "Horned Lizard", "It still hasn't figured out how to stop rolling once it starts.",
        ),
        Entry(
            7, "Cave Mole", "Who's got teeth like spades and needs no shades? CAVE MOLE! CAVE MOLE!",
        ),
        Entry(
            8, "Quillback", "The leader of the cavemen. A big bully who wears the pelt of the last cave porcupine.",
        ),
        Entry(
            9, "Mantrap", "Once it tasted human flesh, it never looked back!",
        ),
        Entry(
            10, "Tiki man", "A masked Jungle-dweller who hunts with a large boomerang.",
        ),
        Entry(
            11, "Witch Docter", "A Tiki Tribe elder who practices black magic. They use an effigy to attack intruders.",
        ),
        Entry(
            12, "Mosquito", "An annoying pest! Makes me mad just thinking about it.",
        ),
        Entry(
            13, "Monkey", "A mischievous forest trickster. It simply can't get enough of its own unfunny pranks!",
        ),
        Entry(
            14, "Hang Spider", "It could chase you if it had to but it'd rather not leave its web if possible.",
        ),
        Entry(
            15, "Giant Spider", "A frightening arachnid the size of a car. She squirts sticky webs to ensnare her prey.",
        ),
        Entry(
            16, "Magmar", "When an adventurer dies in lava, their spirit can become trapped inside the lava instead of going free.",
        ),
        Entry(
            17, "Robot", "They patrol Volcana aimlessly, awaiting instructions from long-dead masters",
        ),
        Entry(
            18, "Fire Bug", "A deadly insect that protects itself by igniting gas stored inside its body.",
        ),
        Entry(
            19, "Imp", "When Yama disappeared, these lowly servants emerged from Hell seeking employment",
        ),
        Entry(
            20, "Lavamander", "The blood of this lizard-like monster runs so cold that only the immense heat of lava can keep it going.",
        ),
        Entry(
            21, "Vampire", "A tortured soul who traded their humanity for eternal life.",
        ),
        Entry(
            22, "Vlad", "The dark lord of all vampires. He relocated to a large castle deep inside Volcana.",
        ),
        Entry(
            23, "Olmec", "Once thought to be the cause of a curse that trapped explorers in a cave... but maybe he was just a small part of it?",
        ),
        Entry(
            24, "Jiangshi", "A reanimated corpse. Its stiff legs force it to hop around.",
        ),
        Entry(
            25, "Jiangshi Assassin", "This wily jiang-shi was once an imperial assassin.",
        ),
        Entry(
            26, "Flying Fish", "A strange, thrill-seeking fish that will go out of its way to eat \"landfood\".",
        ),
        Entry(
            27, "Octopy", "It defends itself with a blinding ink attack and a crown of razor sharp barnacles.",
        ),
        Entry(
            28, "Hermit Crab", "An irritable loner who hates being disturbed. Its saliva is very poisonous!",
        ),
        Entry(
            29, "Pangxie", "One of Kingu's famous armored guards. His claws can crush wood and stone with ease.",
        ),
        Entry(
            30, "Great Humphead", "Endlessly he patrols the salty waters of the Great Lake, waiting for the Empress to return and feed him.",
        ),
        Entry(
            31, "Kingu", "Tiamat's loyal companion and unwavering protector. Her mighty shell once withstood the impact of a comet.",
        ),
        Entry(
            32, "Crocman", "Their powerful bite and teleportation magic makes them a dangerous foe.",
        ),
        Entry(
            33, "Cobra", "A hooded serpent with a rather rude habit.",
        ),
        Entry(
            34, "Mummy", "This frightening tomb lord emerges from its eternal slumber to protect the Temple from trespassers.",
        ),
        Entry(
            35, "Sorceress", "An expert magic user who wields the power of creation",
        ),
        Entry(
            36, "Cat Mummy", "The mummified felines are filled with terrible curses.",
        ),
        Entry(
            37, "Necromancer", "A master of the dark arts who dedicated his life to studying the Book of the Dead.",
        ),
        Entry(
            38, "Anubis", "This legendary jackal-headed god is tasked with guarding the Temple's most hallowed treasures.",
        ),
        Entry(
            39, "Ammit", "A terrifying underworld demon who eats the hearts of unclean souls, casuing them to die a second death.",
        ),
        Entry(
            40, "Apep", "A great snake god who slithers tirelessly throughout Duat, with scales of flint and a stomach full of hot ash.",
        ),
        Entry(
            41, "Anubis II", "Having been defeated at the Temple, Anubis returns to the underworld to beg forgiveness of his father, Osiris.",
        ),
        Entry(
            42, "Osiris", "The ruler of Duat and the ultimate judge of souls who make their way to him.",
        ),
        Entry(
            43, "UFO", "In alien society, being a flying saucer pilot is considered a lowly occupation.",
        ),
        Entry(
            44, "Alien", "Tiamat's firstborn. They harbor a deep jealousy of the Olmites, whom they think she favors.",
        ),
        Entry(
            45, "Yeti", "It loves to toss things and chew gum. But it doesn't know what gum is!",
        ),
        Entry(
            46, "Yeti King", "The king of the yetis is whichever one is biggest.",
        ),
        Entry(
            47, "Yeti Queen", "The queen of the yetis is whichever one can default the king in battle.",
        ),
        Entry(
            48, "Lahamu", "The eldest daughter of Tiamat. A jealous and vindictive warlord with powerful psychic abilities",
        ),
        Entry(
            49, "Proto Shopkeeper", "During the First Wars, Lahamu stole the Original Man to create clones of him with varying degrees of success.",
        ),
        Entry(
            50, "Olmite", "Tiamat's secondborn. An inquisitive race that excels at creation.",
        ),
        Entry(
            51, "Lamassu", "A mighty demigod created by the Olmites to help them guard Neo Babylon.",
        ),
        Entry(
            52, "Tiamat", "The Queen of the World. Though myth and legend hava separated her from her decendents, some still hear her call...",
        ),
        Entry(
            53, "Tadpole", "A baby frog. It's content to splash in the murky waters.",
        ),
        Entry(
            54, "Frog", "Over the aeons, these bloated amphibians have overrun the Sunken City.",
        ),
        Entry(
            55, "Fire Frog", "Some of the Sunken City frogs have developed an explosive personality.",
        ),
        Entry(
            56, "Goliath Frog", "A frog of legend. It carries its entire family within its mighty belly.",
        ),
        Entry(
            57, "Grub", "A lowly creature just trying to make it to the next part of its lifecycle.",
        ),
        Entry(
            58, "Giant Fly", "A fly of titanic proportions. It looks threatening but seems rather uninterested in us...",
        ),
        Entry(
            59, "Hundun", "The muddled dreamer Hundun. The cosmic egg Hunden. The majestic child Hundun. The unstoppable Hunden.",
        ),
        Entry(
            60, "Eggplant Minister", "The most loyal servants of the King of Hell, who continued to follow though commands no longer left his lips.",
        ),
        Entry(
            61, "Eggplup", "Once a fierce hound of hell, now a playful puppy.",
        ),
        Entry(
            62, "Celestial Jelly", "These mighty titan's roam the vast waters of the Cosmic Ocean, leaving behind them great ripples that alter time and space.",
        ),
        Entry(
            63, "Scorpion", "Sometimes you break open a pot and a nightmare comes out. This is that nightmare.",
        ),
        Entry(
            64, "Bee", "An important part of the cave ecosystem. However, it's much larger and more aggressive than the common honeybee.",
        ),
        Entry(
            65, "Queen Bee", "The matriarch of the hive. On her abdomen one might find a blob of delicious royal jelly.",
        ),
        Entry(
            66, "Scarab", "A jeweled flying beetle revered by the followers of Anubis and sought after by leprechauns and treasure hunters.",
        ),
        Entry(
            67, "Golden Monkey", "He struck out on adventure but a sorceress used magic to transform him into an idol.",
        ),
        Entry(
            68, "Leprechaun", "This greedy sprite seeks out heavy pockets to pilfer for his treasure stash.",
        ),
        Entry(
            69, "Monty", "A helpless pug adopted by Tina Flan after her first big adventure. Ana's somewhat faithful companion.",
        ),
        Entry(
            70, "Percy", "A helpless cat who snuck aboard Margaret Tunnel's spaceship looking for a warm place to sleep.",
        ),
        Entry(
            71, "Poochi", "A helpless hamster who Colin Northward brought to the Moon in the name of science. Unfortunately, she escaped.",
        ),
        Entry(
            72, "Ghist", "If in life the spirit is small and spiteful, in death its presence is weak and hard to know.",
        ),
        Entry(
            73, "Ghost", "This angry and relentless spirit chases down explorers for reasons unknown.",
        ),
        Entry(
            74, "Cave Turkey", "A large, meaty bird, best known for its fleshy wattle, also known as a snood.",
        ),
        Entry(
            75, "Rock Dog", "The skin of this fire-breathing canine is protected by a rock-hard plates, hence the name.",
        ),
        Entry(
            76, "Axolotl", "The empress encountered these curious creatures during her travels and brought some back as pets. They quickly escaped!",
        ),
        Entry(
            77, "Qilin", "A creature of myth, whose arrival is said to coincide with important events in the history of humans and gods.",
        ),
        Entry(
            78, "Mech Rider", "A MOYA-class mechanized armor designed by aliens to combat the Olmites during the Second Wars.",
        ),

    ],
    "traps": [
        Entry(
            1, "Spikes", "Ah, spikes! The classic trap. They come in many shapes and sizes, all painful.",
        ),
        Entry(
            2, "Arrow Trap", "An ancient trap of unknown origin. It's responsible for the death of many an explorer.",
        ),
        Entry(
            3, "Bear Trap", "Quillback built these to train his followers, although they mostly serve to put them \"to sleep\".",
        ),
        Entry(
            4, "Log Trap", "The trunk of a massive tree, carved by Quillback and strung up to protect his treasure.",
        ),
        Entry(
            5, "Spear Trap", "These wooden spear traps can be hard to spot amidst the Jungle foliage.",
        ),
        Entry(
            6, "Thorny Trap", "The thorns of these thick vines aren't long enough to impale, but they're sharp enough to deliver a nasty poke.",
        ),
        Entry(
            7, "Snap Trap", "The slightest pressure will cause its powerful steel jaws to snap shut, causing massive damage.",
        ),
        Entry(
            8, "Powder Box", "A wooden box hastily filled with black powder. It will explode at the slightest provocation.",
        ),
        Entry(
            9, "Falling Platform", "A rickety outcropping of rock that won't take very much weight...",
        ),
        Entry(
            10, "Spikeball", "A large ball-and-chain that was designed for crushing rock.",
        ),
        Entry(
            11, "Lion Trap", "It's said that the Empress successfully deflected a barbarian horde using only these golden lions.",
        ),
        Entry(
            12, "Giant Clam", "Over the years, the Empress's treasure has made its way to the ocean and into the mouths of these oversized molluscs.",
        ),
        Entry(
            13, "Sliding Wall", "A door-like mechanism that is activated by a switch.",
        ),
        Entry(
            14, "Crush Trap", "A solid block of stone designed by the pharaohs to flatten would-be tomb robbers.",
        ),
        Entry(
            15, "Giant Crush Trap", "An extra large edition of the crush trap, for discerning pharaohs who want that extra crushing power.",
        ),
        Entry(
            16, "Boulder", "A frosty boulder released to smash treasure hunters that make it to the Ice Caves.",
        ),
        Entry(
            17, "Spring Trap", "In alien society, bouncing up and down is a common form of recreation and exercise.",
        ),
        Entry(
            18, "Landmine", "An explosive device triggered by proximity. Disable it before it's too late!",
        ),
        Entry(
            19, "Laser Trap", "A fixed laser gun triggered by a motion detector.",
        ),
        Entry(
            20, "Spark Trap", "The enormous amount of excess energy generated by Neo Babylon is channeled into dangerous contraptions such as these.",
        ),
        Entry(
            21, "Frog Trap", "The spear is so huge that it's hard to imagine that it was meant to be used against human beings...",
        ),
        Entry(
            22, "Sticky Trap", "A strange growth, whose fleshy protuberance is reminiscent of a uvula.",
        ),
        Entry(
            23, "Bone Drop", "This mass of bones is so large that it's starting to burst out of the ground.",
        ),
        Entry(
            24, "Egg Sac", "Beneath the thin membrane you can see something writhing around vigorously.",
        ),
    ],
}



from struct import pack, unpack
from PIL import Image
import io

from .chacha import decrypt_data, filename_hash
from .assets import KNOWN_ASSETS


class AssetDetails(object):
    def __init__(self, filename, hashed_name, encrypted, offset, size):
        self.filename = filename
        self.hashed_name = hashed_name
        self.encrypted = encrypted
        self.offset = offset
        self.size = size

    def __repr__(self):
        return "AssetDetails(filename={}, encrypted={}, offset={}, size={})".format(
            self.filename, self.encrypted, hex(self.offset), self.size,
        )

    def extract(self, handle):
        handle.seek(self.offset)
        data = handle.read(self.size)
        if self.encrypted:
            data = decrypt_data(self.filename, data)

        if self.filename.endswith(b".png"):
            width, height = unpack('<II', data[:8])
            image = Image.frombytes('RGBA', (width, height), data, 'raw')
            new_data = io.BytesIO()
            image.save(new_data, format="PNG")
            data = new_data.getvalue()

        return data


class Extractor(object):
    def __init__(self, exe_handle):
        self.exe_handle = exe_handle

        self.asset_hashes = {}
        for asset in KNOWN_ASSETS:
            self.asset_hashes[filename_hash(asset)] = asset

    def get_filename(self, hashed_name):
        filename = self.asset_hashes.get(hashed_name)
        if filename:
            return filename

        hashed_len = len(hashed_name)
        for key in self.asset_hashes:
            min_len = min(hashed_len, len(key))
            if hashed_name[:min_len] == key[:min_len]:
                return self.asset_hashes[key]

        return None

    def extract_asset_details(self, offset=0x400):
        self.exe_handle.seek(offset)

        while True:
            asset_len, name_len = unpack('<II', self.exe_handle.read(8))
            if (asset_len, name_len) == (0, 0):
                break
            assert asset_len > 0

            hashed_name = self.exe_handle.read(name_len)
            # Remove NULL byte
            hashed_name = hashed_name[:-1]
            encrypted = self.exe_handle.read(1) == b'\x01'
            offset = self.exe_handle.tell()
            size = asset_len - 1

            self.exe_handle.seek(size, 1)

            filename = self.get_filename(hashed_name)
            if not filename:
                print("Unknown hash. Skipping...")
                continue

            yield AssetDetails(
                filename=filename,
                hashed_name=hashed_name,
                encrypted=encrypted,
                offset=offset,
                size=size,
            )


def main():
    import sys

    if len(sys.argv) < 4:
        print("./s2-asset-extract path/to/bin asset-name dest-name")

    with open(sys.argv[1], 'rb') as exe:
        extractor = Extractor(exe)
        assets = {}
        for asset_details in extractor.extract_asset_details():
            assets[asset_details.filename] = asset_details

        with open(sys.argv[3], "wb") as f:
            f.write(assets[sys.argv[2].encode()].extract(exe))



#!/usr/bin/env python3

from s2_data import field_descriptors
import io
import sys

def to_text():
    with io.open(sys.argv[1], 'rb') as save_file:
        save = save_file.read()

        for category_key, category in field_descriptors.items():
            print('##', category.name, end='\n\n')

            for dkey, descriptor in category.fields.items():
                field_start = descriptor.offset
                field_end = descriptor.offset + descriptor.type.size

                value = descriptor.type.from_binary(save[field_start:field_end])

                print(f'{descriptor.name:30} =', value)

            print(end='\n\n')



"""
Library for extracting assets from Spelunky 2.

The vast majority of the work behind extracting these assets comes
from the work of SciresM.

https://gist.github.com/SciresM/d97e21a02d4a4cdc11b2b97cf43efea3

Additional a lot of clean up and expansion of that work came from
Cloppershy

https://gist.github.com/Cloppershy/a32b9139f3d222b5ff5b8b23ffac1aac

Both of these contributions were crucial to this library.

"""




from abc import ABC, abstractmethod
from enum import IntFlag, IntEnum
import typing

class Type(ABC):
    @property
    @abstractmethod
    def size(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    def from_binary(self, binary: bytes) -> typing.Any:
        raise NotImplementedError()

    @abstractmethod
    def to_binary(self, value: typing.Any) -> bytes:
        raise NotImplementedError()

class BoolType(Type):
    @property
    def size(self):
        return 1

    def from_binary(self, binary):
        return binary[0] != 0

    def to_binary(self, value):
        return bytes([1 if value != 0 else 0])

class Int32Type(Type):
    @property
    def size(self):
        return 4

    def from_binary(self, binary):
        return int.from_bytes(binary, byteorder='little', signed=True)

    def to_binary(self, value):
        return value.to_bytes(length=self.size, byteorder='little', signed=True)

class UInt32Type(Type):
    @property
    def size(self):
        return 4

    def from_binary(self, binary):
        return int.from_bytes(binary, byteorder='little', signed=False)

    def to_binary(self, value):
        return value.to_bytes(length=self.size, byteorder='little', signed=False)

class Int64Type(Type):
    @property
    def size(self):
        return 8

    def from_binary(self, binary):
        return int.from_bytes(binary, byteorder='little', signed=True)

    def to_binary(self, value):
        return value.to_bytes(length=self.size, byteorder='little', signed=True)

class ByteType(Type):
    @property
    def size(self):
        return 1

    def from_binary(self, binary):
        return binary[0]

    def to_binary(self, value):
        return bytes([value])

class FlagType(Type):
    _size = None
    flag_definition = None

    def __init__(self, size, flag_definition):
        self._size = size
        self.flag_definition = flag_definition

    @property
    def size(self):
        return self._size

    def from_binary(self, binary):
        return self.flag_definition(int.from_bytes(binary, byteorder='little'))

    def to_binary(self, value):
        return value.to_bytes(length=self.size, byteorder='little')

class EnumType(Type):
    _size = None
    enum_definition = None

    def __init__(self, size, enum_definition):
        self._size = size
        self.enum_definition = enum_definition

    @property
    def size(self):
        return self._size

    def from_binary(self, binary):
        return self.enum_definition(int.from_bytes(binary, byteorder = 'little'))

    def to_binary(self, value):
        return value.to_bytes(length = self.size, byteorder = 'little')



from struct import pack, unpack
import zstandard as zstd

def rotate_left(a, b):
    return ((((a) << (b)) | ((a) >> (32 - (b))))) & 0xFFFFFFFF

def quarter_round(w, a, b, c, d):
    w[a] += w[b]
    w[a] &= 0xFFFFFFFF
    w[d] ^= w[a]
    w[d] = rotate_left(w[d], 16)
    w[c] += w[d]
    w[c] &= 0xFFFFFFFF
    w[b] ^= w[c]
    w[b] = rotate_left(w[b], 12)
    w[a] += w[b]
    w[a] &= 0xFFFFFFFF
    w[d] ^= w[a]
    w[d] = rotate_left(w[d],  8)
    w[c] += w[d]
    w[c] &= 0xFFFFFFFF
    w[b] ^= w[c]
    w[b] = rotate_left(w[b],  7)

def round_pair(w):
    quarter_round(w,  0,  4,  8, 12)
    quarter_round(w,  1,  5,  9, 13)
    quarter_round(w,  2,  6, 10, 14)
    quarter_round(w,  3,  7, 11, 15)
    quarter_round(w,  0,  5, 10, 15)
    quarter_round(w,  1,  6, 11, 12)
    quarter_round(w,  2,  7,  8, 13)
    quarter_round(w,  3,  4,  9, 14)

def two_rounds(s):
    w = s_to_w(s)
    round_pair(w)
    round_pair(w)
    return w_to_s(w)

def quad_rounds(s):
    w = s_to_w(s)
    round_pair(w)
    round_pair(w)
    round_pair(w)
    round_pair(w)
    return w_to_s(w)

def sxor(x, y):
    return bytes(bytearray(a ^ b for a, b in zip(bytearray(x), bytearray(y))))

def s_to_w(s):
    return list(unpack(b'<'+(b'I'*(len(s)//4)), s))

def w_to_s(w):
    return pack(b'<'+(b'I'*len(w)), *w)

def s_to_q(s):
    return list(unpack(b'<'+(b'Q'*(len(s)//8)), s))

def q_to_s(w):
    return pack(b'<'+(b'Q'*len(w)), *w)

def add_qwords(h0, h1):
    return q_to_s([(a + b) & 0xFFFFFFFFFFFFFFFF for a,b in zip(s_to_q(h0), s_to_q(h1))])

def mix_in(h, s):
    def mix_partial(h, partial):
        assert len(partial) <= 0x40
        b = bytearray(h)
        for i, c in enumerate(partial[::-1]):
            b[i] ^= ord('%c' % c)
        return quad_rounds(bytes(b))

    while s != b'':
        h = mix_partial(h, s[:0x40])
        s = s[0x40:]

    return h

def filename_hash(s):
    # Generate initial hash from the string
    h0 = mix_in(b'\x00'*0x40, s)

    # Advance h0 by four round pairs to get h1
    h1 = quad_rounds(h0)
    # Add the two together, and advance by four round pairs.
    key = quad_rounds(add_qwords(h0, h1))

    # Do keyed hashing
    # NOTE: This appears to be an implementation mistake on the Spelunky 2 dev's part
    # They generate a quad_round advanced version of (nonce'd key), but then they
    # xor with the untweaked key instead of the tweaked key...
    h = b''
    for i in range(0, len(s), 0x40):
        partial = s[i:i+0x40]
        h += sxor(partial, key[:len(partial)][::-1])
    return h

def decrypt_data(s, data):
    # Untweaked key begins as half-advanced "0xBABE"
    h = two_rounds(pack(b'<QQQQQQQQ', 0xBABE, 0, 0, 0, 0, 0, 0, 0))

    # Mix the filename in to tweak the key
    for i in range(0, len(s), 0x40):
        partial = s[i:i+0x40]
        h = quad_rounds(sxor(h[:len(partial)], partial[::-1]) + h[len(partial):])

    # Add the tweaked key and its advancement, then advance by four round pairs.
    key = quad_rounds(add_qwords(h, quad_rounds(h)))

    # NOTE: This appears to be an implementation mistake on the Spelunky 2 dev's part
    # They generate a quad_round advanced version of (nonce'd key), but then they
    # xor with the untweaked key instead of the tweaked key...
    dec = b''
    if len(data) >= 0x40:
        blocks = len(data) // 0x40
        dec += sxor(data, key[::-1] * blocks)
        data = data[blocks * 0x40:]
    if len(data) > 0:
        dec += sxor(data, key[:len(data)][::-1])

    cctx = zstd.ZstdDecompressor()
    decompressed = cctx.decompress(dec)
    return decompressed



from .field_types import BoolType, ByteType, Int32Type, UInt32Type, Int64Type, FlagType, EnumType
from .field_descriptors import field_descriptors, CharacterUnlockFlags, ShortcutProgressEnum, CharacterSelectedEnum



KNOWN_ASSETS = [
    b"Data/Fonts/fontdebug.fnb",
    b"Data/Fonts/fontfirasans.fnb",
    b"Data/Fonts/fontmono.fnb",
    b"Data/Fonts/fontyorkten.fnb",
    b"Data/Levels/abzu.lvl",
    b"Data/Levels/Arena/dm1-1.lvl",
    b"Data/Levels/Arena/dm1-2.lvl",
    b"Data/Levels/Arena/dm1-3.lvl",
    b"Data/Levels/Arena/dm1-4.lvl",
    b"Data/Levels/Arena/dm1-5.lvl",
    b"Data/Levels/Arena/dm2-1.lvl",
    b"Data/Levels/Arena/dm2-2.lvl",
    b"Data/Levels/Arena/dm2-3.lvl",
    b"Data/Levels/Arena/dm2-4.lvl",
    b"Data/Levels/Arena/dm2-5.lvl",
    b"Data/Levels/Arena/dm3-1.lvl",
    b"Data/Levels/Arena/dm3-2.lvl",
    b"Data/Levels/Arena/dm3-3.lvl",
    b"Data/Levels/Arena/dm3-4.lvl",
    b"Data/Levels/Arena/dm3-5.lvl",
    b"Data/Levels/Arena/dm4-1.lvl",
    b"Data/Levels/Arena/dm4-2.lvl",
    b"Data/Levels/Arena/dm4-3.lvl",
    b"Data/Levels/Arena/dm4-4.lvl",
    b"Data/Levels/Arena/dm4-5.lvl",
    b"Data/Levels/Arena/dm5-1.lvl",
    b"Data/Levels/Arena/dm5-2.lvl",
    b"Data/Levels/Arena/dm5-3.lvl",
    b"Data/Levels/Arena/dm5-4.lvl",
    b"Data/Levels/Arena/dm5-5.lvl",
    b"Data/Levels/Arena/dm6-1.lvl",
    b"Data/Levels/Arena/dm6-2.lvl",
    b"Data/Levels/Arena/dm6-3.lvl",
    b"Data/Levels/Arena/dm6-4.lvl",
    b"Data/Levels/Arena/dm6-5.lvl",
    b"Data/Levels/Arena/dm7-1.lvl",
    b"Data/Levels/Arena/dm7-2.lvl",
    b"Data/Levels/Arena/dm7-3.lvl",
    b"Data/Levels/Arena/dm7-4.lvl",
    b"Data/Levels/Arena/dm7-5.lvl",
    b"Data/Levels/Arena/dm8-1.lvl",
    b"Data/Levels/Arena/dm8-2.lvl",
    b"Data/Levels/Arena/dm8-3.lvl",
    b"Data/Levels/Arena/dm8-4.lvl",
    b"Data/Levels/Arena/dm8-5.lvl",
    b"Data/Levels/Arena/dmpreview.tok",
    b"Data/Levels/babylonarea.lvl",
    b"Data/Levels/basecamp_garden.lvl",
    b"Data/Levels/basecamp.lvl",
    b"Data/Levels/basecamp_shortcut_discovered.lvl",
    b"Data/Levels/basecamp_shortcut_undiscovered.lvl",
    b"Data/Levels/basecamp_shortcut_unlocked.lvl",
    b"Data/Levels/basecamp_surface.lvl",
    b"Data/Levels/basecamp_tutorial.lvl",
    b"Data/Levels/basecamp_tv_room_locked.lvl",
    b"Data/Levels/basecamp_tv_room_unlocked.lvl",
    b"Data/Levels/beehive.lvl",
    b"Data/Levels/blackmarket.lvl",
    b"Data/Levels/cavebossarea.lvl",
    b"Data/Levels/challenge_moon.lvl",
    b"Data/Levels/challenge_star.lvl",
    b"Data/Levels/challenge_sun.lvl",
    b"Data/Levels/cityofgold.lvl",
    b"Data/Levels/cosmicocean_babylon.lvl",
    b"Data/Levels/cosmicocean_dwelling.lvl",
    b"Data/Levels/cosmicocean_icecavesarea.lvl",
    b"Data/Levels/cosmicocean_jungle.lvl",
    b"Data/Levels/cosmicocean_sunkencity.lvl",
    b"Data/Levels/cosmicocean_temple.lvl",
    b"Data/Levels/cosmicocean_tidepool.lvl",
    b"Data/Levels/cosmicocean_volcano.lvl",
    b"Data/Levels/duat.lvl",
    b"Data/Levels/dwellingarea.lvl",
    b"Data/Levels/eggplantarea.lvl",
    b"Data/Levels/ending.lvl",
    b"Data/Levels/generic.lvl",
    b"Data/Levels/hallofushabti.lvl",
    b"Data/Levels/hundun.lvl",
    b"Data/Levels/icecavesarea.lvl",
    b"Data/Levels/junglearea.lvl",
    b"Data/Levels/lake.lvl",
    b"Data/Levels/lakeoffire.lvl",
    b"Data/Levels/olmecarea.lvl",
    b"Data/Levels/palaceofpleasure.lvl",
    b"Data/Levels/sunkencityarea.lvl",
    b"Data/Levels/templearea.lvl",
    b"Data/Levels/testingarea.lvl",
    b"Data/Levels/tiamat.lvl",
    b"Data/Levels/tidepoolarea.lvl",
    b"Data/Levels/vladscastle.lvl",
    b"Data/Levels/volcanoarea.lvl",
    b"Data/Textures/base_eggship2.png",
    b"Data/Textures/base_eggship3.png",
    b"Data/Textures/base_eggship.png",
    b"Data/Textures/base_skynight.png",
    b"Data/Textures/base_surface2.png",
    b"Data/Textures/base_surface.png",
    b"Data/Textures/bayer8.png",
    b"Data/Textures/bg_babylon.png",
    b"Data/Textures/bg_beehive.png",
    b"Data/Textures/bg_cave.png",
    b"Data/Textures/bg_duat2.png",
    b"Data/Textures/bg_duat.png",
    b"Data/Textures/bg_eggplant.png",
    b"Data/Textures/bg_gold.png",
    b"Data/Textures/bg_ice.png",
    b"Data/Textures/bg_jungle.png",
    b"Data/Textures/bg_mothership.png",
    b"Data/Textures/bg_stone.png",
    b"Data/Textures/bg_sunken.png",
    b"Data/Textures/bg_temple.png",
    b"Data/Textures/bg_tidepool.png",
    b"Data/Textures/bg_vlad.png",
    b"Data/Textures/bg_volcano.png",
    b"Data/Textures/border_main.png",
    b"Data/Textures/char_black.png",
    b"Data/Textures/char_blue.png",
    b"Data/Textures/char_cerulean.png",
    b"Data/Textures/char_cinnabar.png",
    b"Data/Textures/char_cyan.png",
    b"Data/Textures/char_eggchild.png",
    b"Data/Textures/char_gold.png",
    b"Data/Textures/char_gray.png",
    b"Data/Textures/char_green.png",
    b"Data/Textures/char_hired.png",
    b"Data/Textures/char_iris.png",
    b"Data/Textures/char_khaki.png",
    b"Data/Textures/char_lemon.png",
    b"Data/Textures/char_lime.png",
    b"Data/Textures/char_magenta.png",
    b"Data/Textures/char_olive.png",
    b"Data/Textures/char_orange.png",
    b"Data/Textures/char_pink.png",
    b"Data/Textures/char_red.png",
    b"Data/Textures/char_violet.png",
    b"Data/Textures/char_white.png",
    b"Data/Textures/char_yellow.png",
    b"Data/Textures/coffins.png",
    b"Data/Textures/credits.png",
    b"Data/Textures/deco_babylon.png",
    b"Data/Textures/deco_basecamp.png",
    b"Data/Textures/deco_cave.png",
    b"Data/Textures/deco_cosmic.png",
    b"Data/Textures/deco_eggplant.png",
    b"Data/Textures/deco_extra.png",
    b"Data/Textures/deco_gold.png",
    b"Data/Textures/deco_ice.png",
    b"Data/Textures/deco_jungle.png",
    b"Data/Textures/deco_sunken.png",
    b"Data/Textures/deco_temple.png",
    b"Data/Textures/deco_tidepool.png",
    b"Data/Textures/deco_tutorial.png",
    b"Data/Textures/deco_volcano.png",
    b"Data/Textures/floor_babylon.png",
    b"Data/Textures/floor_cave.png",
    b"Data/Textures/floor_eggplant.png",
    b"Data/Textures/floor_ice.png",
    b"Data/Textures/floor_jungle.png",
    b"Data/Textures/floormisc.png",
    b"Data/Textures/floorstyled_babylon.png",
    b"Data/Textures/floorstyled_beehive.png",
    b"Data/Textures/floorstyled_duat.png",
    b"Data/Textures/floorstyled_gold_normal.png",
    b"Data/Textures/floorstyled_gold.png",
    b"Data/Textures/floorstyled_guts.png",
    b"Data/Textures/floorstyled_mothership.png",
    b"Data/Textures/floorstyled_pagoda.png",
    b"Data/Textures/floorstyled_palace.png",
    b"Data/Textures/floorstyled_stone.png",
    b"Data/Textures/floorstyled_sunken.png",
    b"Data/Textures/floorstyled_temple.png",
    b"Data/Textures/floorstyled_vlad.png",
    b"Data/Textures/floorstyled_wood.png",
    b"Data/Textures/floor_sunken.png",
    b"Data/Textures/floor_surface.png",
    b"Data/Textures/floor_temple.png",
    b"Data/Textures/floor_tidepool.png",
    b"Data/Textures/floor_volcano.png",
    b"Data/Textures/fontdebug.png",
    b"Data/Textures/fontfirasans.png",
    b"Data/Textures/fontmono.png",
    b"Data/Textures/fontyorkten.png",
    b"Data/Textures/fx_ankh.png",
    b"Data/Textures/fx_big.png",
    b"Data/Textures/fx_explosion.png",
    b"Data/Textures/fx_rubble.png",
    b"Data/Textures/fx_small2.png",
    b"Data/Textures/fx_small3.png",
    b"Data/Textures/fx_small.png",
    b"Data/Textures/hud_controller_buttons.png",
    b"Data/Textures/hud.png",
    b"Data/Textures/hud_text.png",
    b"Data/Textures/items.png",
    b"Data/Textures/items_ushabti.png",
    b"Data/Textures/journal_back.png",
    b"Data/Textures/journal_elements.png",
    b"Data/Textures/journal_entry_bg.png",
    b"Data/Textures/journal_entry_items.png",
    b"Data/Textures/journal_entry_mons_big.png",
    b"Data/Textures/journal_entry_mons.png",
    b"Data/Textures/journal_entry_people.png",
    b"Data/Textures/journal_entry_place.png",
    b"Data/Textures/journal_entry_traps.png",
    b"Data/Textures/journal_pageflip.png",
    b"Data/Textures/journal_pagetorn.png",
    b"Data/Textures/journal_select.png",
    b"Data/Textures/journal_stickers.png",
    b"Data/Textures/journal_story.png",
    b"Data/Textures/journal_top_entry.png",
    b"Data/Textures/journal_top_gameover.png",
    b"Data/Textures/journal_top_main.png",
    b"Data/Textures/journal_top_profile.png",
    b"Data/Textures/loading.png",
    b"Data/Textures/lut_backlayer.png",
    b"Data/Textures/lut_blackmarket.png",
    b"Data/Textures/lut_icecaves.png",
    b"Data/Textures/lut_original.png",
    b"Data/Textures/lut_vlad.png",
    b"Data/Textures/main_body.png",
    b"Data/Textures/main_dirt.png",
    b"Data/Textures/main_doorback.png",
    b"Data/Textures/main_doorframe.png",
    b"Data/Textures/main_door.png",
    b"Data/Textures/main_fore1.png",
    b"Data/Textures/main_fore2.png",
    b"Data/Textures/main_head.png",
    b"Data/Textures/menu_basic.png",
    b"Data/Textures/menu_brick1.png",
    b"Data/Textures/menu_brick2.png",
    b"Data/Textures/menu_cave1.png",
    b"Data/Textures/menu_cave2.png",
    b"Data/Textures/menu_chardoor.png",
    b"Data/Textures/menu_charsel.png",
    b"Data/Textures/menu_deathmatch2.png",
    b"Data/Textures/menu_deathmatch3.png",
    b"Data/Textures/menu_deathmatch4.png",
    b"Data/Textures/menu_deathmatch5.png",
    b"Data/Textures/menu_deathmatch6.png",
    b"Data/Textures/menu_deathmatch.png",
    b"Data/Textures/menu_disp.png",
    b"Data/Textures/menu_generic.png",
    b"Data/Textures/menu_header.png",
    b"Data/Textures/menu_leader.png",
    b"Data/Textures/menu_online.png",
    b"Data/Textures/menu_titlegal.png",
    b"Data/Textures/menu_title.png",
    b"Data/Textures/menu_tunnel.png",
    b"Data/Textures/monsters01.png",
    b"Data/Textures/monsters02.png",
    b"Data/Textures/monsters03.png",
    b"Data/Textures/monstersbasic01.png",
    b"Data/Textures/monstersbasic02.png",
    b"Data/Textures/monstersbasic03.png",
    b"Data/Textures/monstersbig01.png",
    b"Data/Textures/monstersbig02.png",
    b"Data/Textures/monstersbig03.png",
    b"Data/Textures/monstersbig04.png",
    b"Data/Textures/monstersbig05.png",
    b"Data/Textures/monstersbig06.png",
    b"Data/Textures/monsters_ghost.png",
    b"Data/Textures/monsters_hundun.png",
    b"Data/Textures/monsters_olmec.png",
    b"Data/Textures/monsters_osiris.png",
    b"Data/Textures/monsters_pets.png",
    b"Data/Textures/monsters_tiamat.png",
    b"Data/Textures/monsters_yama.png",
    b"Data/Textures/mounts.png",
    b"Data/Textures/noise0.png",
    b"Data/Textures/noise1.png",
    b"Data/Textures/OldTextures/ai.png",
    b"Data/Textures/saving.png",
    b"Data/Textures/shadows.png",
    b"Data/Textures/shine.png",
    b"Data/Textures/splash0.png",
    b"Data/Textures/splash1.png",
    b"Data/Textures/splash2.png",
    b"shaders.hlsl",
    b"soundbank.bank",
    b"soundbank.strings.bank",
    b"strings00.str",
    b"strings01.str",
    b"strings02.str",
    b"strings03.str",
    b"strings04.str",
    b"strings05.str",
    b"strings06.str",
    b"strings07.str",
]



from setuptools import setup, find_packages
import pathlib

HERE = pathlib.Path(__file__).parent.resolve()
LONG_DESCRIPTION = (HERE / 'README.md').read_text(encoding='utf-8')
VERSION = (HERE / 'VERSION').read_text(encoding='utf-8').strip()


setup(
    name='s2-data',
    version=VERSION,
    description='Spelunky 2 Data Utilities.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Other Audience',

        # Pick your license as you wish
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],

    keywords='games',
    package_dir={'': 'src'},  # Optional
    packages=find_packages(where='src'),  # Required
    python_requires='>=3.7, <4',

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "zstandard",
        "Pillow",
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    #extras_require={  # Optional
    #    'dev': ['check-manifest'],
    #    'test': ['coverage'],
    #},

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # `pip` to create the appropriate form of executable for the target
    # platform.
    #
    # For example, the following would provide a command called `sample` which
    # executes the function `main` from this package when invoked:
    entry_points={  # Optional
        'console_scripts': [
            's2save-to-text=s2_data.command_line:to_text',
            's2-asset-extract=s2_data.assets.extractor:main',
        ],
    },
    include_package_data = True,
)
